"""Defines the pipeline which handles pipeline elements for molecular operations."""

from collections.abc import Iterable
from typing import Any, Self

import numpy as np
from joblib import Parallel, delayed
from rdkit.Chem.rdchem import MolSanitizeException
from rdkit.rdBase import BlockLogs

from molpipeline.abstract_pipeline_elements.core import (
    ABCPipelineElement,
    InvalidInstance,
    RemovedInstance,
    TransformingPipelineElement,
)
from molpipeline.error_handling import (
    ErrorFilter,
    _MultipleErrorFilter,
)
from molpipeline.utils.molpipeline_types import TypeFixedVarSeq
from molpipeline.utils.multi_proc import check_available_cores


class _MolPipeline:
    """Contains the PipeElements which describe the functionality of the pipeline."""

    _n_jobs: int
    _element_list: list[ABCPipelineElement]
    _requires_fitting: bool

    def __init__(
        self,
        element_list: list[ABCPipelineElement],
        n_jobs: int = 1,
        name: str = "MolPipeline",
    ) -> None:
        """Initialize MolPipeline.

        Parameters
        ----------
        element_list: list[ABCPipelineElement]
            List of Pipeline Elements which form the pipeline.
        n_jobs:
            Number of cores used.
        name: str
            Name of pipeline.

        """
        self._element_list = element_list
        self.n_jobs = n_jobs
        self.name = name
        self._requires_fitting = any(
            element.requires_fitting for element in self._element_list
        )

    @property
    def _filter_elements(self) -> list[ErrorFilter]:
        """Get the elements which filter the input."""
        return [
            element
            for element in self._element_list
            if isinstance(element, ErrorFilter)
        ]

    @property
    def _filter_elements_agg(self) -> _MultipleErrorFilter:
        """Get the aggregated filter element."""
        return _MultipleErrorFilter(self._filter_elements)

    @property
    def _transforming_elements(
        self,
    ) -> list[TransformingPipelineElement | Self]:
        """Get the elements which transform the input."""
        return [
            element
            for element in self._element_list
            if isinstance(element, (TransformingPipelineElement, _MolPipeline))
        ]

    @property
    def n_jobs(self) -> int:
        """Return the number of cores to use in transformation step."""
        return self._n_jobs

    @n_jobs.setter
    def n_jobs(self, requested_jobs: int) -> None:
        """Set the number of cores to use in transformation step.

        Parameters
        ----------
        requested_jobs: int
            Number of cores requested for transformation steps.
            If fewer cores than requested are available, the number of cores is set to
            maximum available.

        """
        self._n_jobs = check_available_cores(requested_jobs)

    @property
    def parameters(self) -> dict[str, Any]:
        """Get all parameters defining the object."""
        return self.get_params()

    @parameters.setter
    def parameters(self, parameter_dict: dict[str, Any]) -> None:
        """Set parameters of the pipeline and pipeline elements.

        Parameters
        ----------
        parameter_dict: dict[str, Any]
            Parameter to be set.

        """
        self.set_params(**parameter_dict)

    @property
    def requires_fitting(self) -> bool:
        """Return whether the pipeline requires fitting."""
        return self._requires_fitting

    def get_params(self, deep: bool = True) -> dict[str, Any]:
        """Get all parameters defining the object.

        Parameters
        ----------
        deep: bool
            If True get a deep copy of the parameters.

        Returns
        -------
        dict[str, Any]
            Dictionary containing the parameter names and corresponding values.

        """
        if deep:
            return {
                "element_list": self.element_list,
                "n_jobs": self.n_jobs,
                "name": self.name,
            }
        return {
            "element_list": self._element_list,
            "n_jobs": self.n_jobs,
            "name": self.name,
        }

    def set_params(self, **parameter_dict: Any) -> Self:
        """Set parameters of the pipeline and pipeline elements.

        Parameters
        ----------
        parameter_dict: Any
            Parameters to be set.

        Returns
        -------
        Self
            MolPipeline object with updated parameters.

        """
        if "element_list" in parameter_dict:
            self._element_list = parameter_dict["element_list"]
        if "n_jobs" in parameter_dict:
            self.n_jobs = int(parameter_dict["n_jobs"])
        if "name" in parameter_dict:
            self.name = str(parameter_dict["name"])
        return self

    @property
    def element_list(self) -> list[ABCPipelineElement]:
        """Get a shallow copy from the list of pipeline elements."""
        return self._element_list[:]  # [:] to create shallow copy.

    def _get_meta_element_list(
        self,
    ) -> "list[ABCPipelineElement | _MolPipeline]":
        """Merge elements which do not require fitting to a meta element.

        This improves parallelization of the pipeline.

        Returns
        -------
        list[ABCPipelineElement | "_MolPipeline"]
            List of pipeline elements and meta elements.

        """
        meta_element_list: list[ABCPipelineElement | _MolPipeline] = []
        no_fit_element_list: list[ABCPipelineElement] = []
        for element in self._element_list:
            if (
                isinstance(element, TransformingPipelineElement)
                and not element.requires_fitting
            ):
                no_fit_element_list.append(element)
            else:
                if len(no_fit_element_list) == 1:
                    meta_element_list.append(no_fit_element_list[0])
                elif len(no_fit_element_list) > 1:
                    meta_element_list.append(
                        _MolPipeline(no_fit_element_list, n_jobs=self.n_jobs),
                    )
                no_fit_element_list = []
                meta_element_list.append(element)
        if len(no_fit_element_list) == 1:
            meta_element_list.append(no_fit_element_list[0])
        elif len(no_fit_element_list) > 1:
            meta_element_list.append(
                _MolPipeline(no_fit_element_list, n_jobs=self.n_jobs),
            )
        return meta_element_list

    def fit(
        self,
        x_input: Any,
        y: Any = None,
        **fit_params: dict[Any, Any],
    ) -> Self:
        """Fit the MolPipeline according to x_input.

        Parameters
        ----------
        x_input: Any
            Molecular representations which are subsequently processed.
        y: Any
            Optional label of input. Only for SKlearn compatibility.
        fit_params: Any
            Parameters. Only for SKlearn compatibility.

        Returns
        -------
        Self
            Fitted MolPipeline.

        """
        _ = y  # Making pylint happy
        _ = fit_params  # Making pylint happy
        if self.requires_fitting:
            self.fit_transform(x_input)
        return self

    def fit_transform(  # pylint: disable=unused-argument
        self,
        x_input: Any,
        y: Any = None,  # noqa: ARG002
        **fit_params: dict[str, Any],  # noqa: ARG002
    ) -> Any:
        """Fit the MolPipeline according to input and return the transformed molecules.

        Parameters
        ----------
        x_input: Any
            Molecular representations which are subsequently processed.
        y: Any
            Optional label of input. Only for SKlearn compatibility.
        fit_params: Any
            Parameters. Only for SKlearn compatibility.

        Returns
        -------
        Any
            Transformed molecules.

        Raises
        ------
        AssertionError
            If a subpipeline requires fitting, which by definition should not be the
            case.

        """
        iter_input = x_input

        removed_rows: dict[ErrorFilter, list[int]] = {}
        for error_filter in self._filter_elements:
            removed_rows[error_filter] = []
        iter_idx_array = np.arange(len(iter_input))

        # The meta elements merge steps which do not require fitting to improve
        # parallelization
        for i_element in self._get_meta_element_list():
            if not isinstance(i_element, (TransformingPipelineElement, _MolPipeline)):
                continue
            i_element.n_jobs = self.n_jobs
            iter_input = i_element.pretransform(iter_input)
            for error_filter in self._filter_elements:
                iter_input = error_filter.transform(iter_input)
                for idx in error_filter.error_indices:
                    idx = iter_idx_array[idx]
                    removed_rows[error_filter].append(idx)
                iter_idx_array = error_filter.co_transform(iter_idx_array)
            if i_element.requires_fitting:
                if isinstance(i_element, _MolPipeline):
                    raise AssertionError("No subpipline should require fitting!")
                i_element.fit_to_result(iter_input)
            if isinstance(i_element, TransformingPipelineElement):
                iter_input = i_element.finalize_list(iter_input)
            iter_input = i_element.assemble_output(iter_input)
            i_element.n_jobs = 1

        # Set removed rows to filter elements to allow for correct co_transform
        iter_idx_array = np.arange(len(x_input))
        for error_filter in self._filter_elements:
            removed_idx_list = removed_rows[error_filter]
            error_filter.error_indices = []
            for new_idx, idx_ in enumerate(iter_idx_array):
                if idx_ in removed_idx_list:
                    error_filter.error_indices.append(new_idx)
            error_filter.n_total = len(iter_idx_array)
            iter_idx_array = error_filter.co_transform(iter_idx_array)
        return iter_input

    def transform_single(self, input_value: Any) -> Any:
        """Transform a single input according to the sequence of PipelineElements.

        Parameters
        ----------
        input_value: Any
            Molecular representation which is subsequently transformed.

        Returns
        -------
        Any
            Transformed molecular representation.

        """
        log_block = BlockLogs()
        iter_value = input_value
        for p_element in self._element_list:
            try:
                if isinstance(iter_value, RemovedInstance):
                    return iter_value
                iter_value = p_element.transform_single(iter_value)
            except MolSanitizeException as err:
                iter_value = InvalidInstance(
                    p_element.uuid,
                    f"RDKit MolSanitizeException: {err.args}",
                    p_element.name,
                )
        del log_block
        return iter_value

    def pretransform(self, x_input: Any) -> Any:
        """Transform the input according to the sequence without assembling the output.

        Parameters
        ----------
        x_input: Any
            Molecular representations which are subsequently transformed.

        Returns
        -------
        Any
            Transformed molecular representations.

        """
        return list(self._transform_iterator(x_input))

    def transform(self, x_input: Any) -> Any:
        """Transform the input according to the sequence of provided PipelineElements.

        Parameters
        ----------
        x_input: Any
            Molecular representations which are subsequently transformed.

        Returns
        -------
        Any
            Transformed molecular representations.

        """
        output_generator = self._transform_iterator(x_input)
        return self.assemble_output(output_generator)

    def assemble_output(self, value_list: Iterable[Any]) -> Any:
        """Assemble the output of the pipeline.

        Parameters
        ----------
        value_list: Iterable[Any]
            Generator which yields the output of the pipeline.

        Returns
        -------
        Any
            Assembled output.

        """
        last_element = self._transforming_elements[-1]
        if hasattr(last_element, "assemble_output"):
            return last_element.assemble_output(value_list)
        return list(value_list)

    def _transform_iterator(self, x_input: Any) -> Any:
        """Transform the input according to the sequence of provided PipelineElements.

        Parameters
        ----------
        x_input: Any
            Molecular representations which are subsequently transformed.

        Yields
        ------
        Any
            Transformed molecular representations.

        """
        agg_filter = self._filter_elements_agg
        for filter_element in self._filter_elements:
            filter_element.error_indices = []
        parallel = Parallel(
            n_jobs=self.n_jobs,
            return_as="generator",
            batch_size="auto",
        )
        output_generator = parallel(
            delayed(self.transform_single)(value) for value in x_input
        )
        for i, transformed_value in enumerate(output_generator):
            if isinstance(transformed_value, RemovedInstance):
                agg_filter.register_removed(i, transformed_value)
            else:
                yield transformed_value
        agg_filter.set_total(len(x_input))

    def co_transform(self, x_input: TypeFixedVarSeq) -> TypeFixedVarSeq:
        """Filter flagged rows from the input.

        Parameters
        ----------
        x_input: Any
            Molecular representations which are subsequently filtered.

        Returns
        -------
        Any
            Filtered molecular representations.

        """
        return self._filter_elements_agg.co_transform(x_input)



"""Unittest for MolFilters functionality.

MolFilters flag Molecules as invalid based on the criteria defined in the filter.

"""

import json
import tempfile
import unittest
from pathlib import Path
from typing import TYPE_CHECKING, Any

import pandas as pd

from molpipeline import ErrorFilter, FilterReinserter, Pipeline
from molpipeline.abstract_pipeline_elements.core import InvalidInstance
from molpipeline.any2mol import SmilesToMol
from molpipeline.mol2any import MolToSmiles
from molpipeline.mol2mol import (
    ComplexFilter,
    ElementFilter,
    InorganicsFilter,
    MixtureFilter,
    RDKitDescriptorsFilter,
    SmartsFilter,
    SmilesFilter,
)
from molpipeline.utils.comparison import compare_recursive
from molpipeline.utils.json_operations import recursive_from_json, recursive_to_json

if TYPE_CHECKING:
    from molpipeline.utils.molpipeline_types import FloatCountRange, IntOrIntCountRange

# pylint: disable=duplicate-code  # test case molecules are allowed to be duplicated
SMILES_ANTIMONY = "[SbH6+3]"
SMILES_BENZENE = "c1ccccc1"
SMILES_CHLOROBENZENE = "Clc1ccccc1"
SMILES_CL_BR = "NC(Cl)(Br)C(=O)O"
SMILES_METAL_AU = "OC[C@H]1OC([S][Au])[C@H](O)[C@@H](O)[C@@H]1O"

SMILES_LIST = [
    SMILES_ANTIMONY,
    SMILES_BENZENE,
    SMILES_CHLOROBENZENE,
    SMILES_METAL_AU,
    SMILES_CL_BR,
]


class ElementFilterTest(unittest.TestCase):
    """Unittest for Elementiflter."""

    def test_element_filter(self) -> None:
        """Test if molecules are filtered correctly by allowed chemical elements."""
        default_atoms_dict = {
            1: (0, None),
            5: (0, None),
            6: (0, None),
            7: (0, None),
            8: (0, None),
            9: (0, None),
            14: (0, None),
            15: (0, None),
            16: (0, None),
            17: (0, None),
            34: (0, None),
            35: (0, None),
            53: (0, None),
        }

        element_filter = ElementFilter()
        self.assertEqual(element_filter.allowed_element_numbers, default_atoms_dict)
        pipeline = Pipeline(
            [
                ("Smiles2Mol", SmilesToMol()),
                ("ElementFilter", element_filter),
                ("Mol2Smiles", MolToSmiles()),
                ("ErrorFilter", ErrorFilter()),
            ],
        )

        test_params_list_with_results = [
            {
                "params": {},
                "result": [SMILES_BENZENE, SMILES_CHLOROBENZENE, SMILES_CL_BR],
            },
            {
                "params": {
                    "ElementFilter__allowed_element_numbers": {
                        6: 6,
                        1: (5, 6),
                        17: (0, 1),
                    },
                },
                "result": [SMILES_BENZENE, SMILES_CHLOROBENZENE],
            },
            {"params": {"ElementFilter__add_hydrogens": False}, "result": []},
        ]

        for test_params in test_params_list_with_results:
            pipeline.set_params(**test_params["params"])
            filtered_smiles = pipeline.fit_transform(SMILES_LIST)
            self.assertEqual(filtered_smiles, test_params["result"])

    def test_json_roundtrip(self) -> None:
        """Test if ElementFilter can be serialized and deserialized.

        Notes
        -----
        It is important to save the ElementFilter as a JSON file and then load it back.
        This is because json.dumps() sets the keys of the dictionary to strings.

        """
        element_filter = ElementFilter()
        json_object = recursive_to_json(element_filter)
        with tempfile.TemporaryDirectory() as temp_folder:
            temp_file_path = Path(temp_folder) / "test.json"
            with temp_file_path.open("w", encoding="UTF-8") as out_file:
                json.dump(json_object, out_file)
            with temp_file_path.open(encoding="UTF-8") as in_file:
                loaded_json_object = json.load(in_file)
        recreated_element_filter = recursive_from_json(loaded_json_object)

        original_params = element_filter.get_params()
        recreated_params = recreated_element_filter.get_params()
        self.assertEqual(original_params.keys(), recreated_params.keys())
        for param_name, original_value in original_params.items():
            with self.subTest(param_name=param_name):
                self.assertTrue(
                    compare_recursive(original_value, recreated_params[param_name]),
                    f"Original: {original_value}, "
                    f"Recreated: {recreated_params[param_name]}",
                )


class ComplexFilterTest(unittest.TestCase):
    """Unittest for ComplexFilter."""

    @staticmethod
    def _create_pipeline() -> Pipeline:
        """Create a pipeline with a complex filter.

        Returns
        -------
        Pipeline
            Pipeline with a complex filter.

        """
        element_filter_1 = ElementFilter({6: 6, 1: 6})
        element_filter_2 = ElementFilter({6: 6, 1: 5, 17: 1})

        multi_element_filter = ComplexFilter(
            (
                ("element_filter_1", element_filter_1),
                ("element_filter_2", element_filter_2),
            ),
        )

        return Pipeline(
            [
                ("Smiles2Mol", SmilesToMol()),
                ("MultiElementFilter", multi_element_filter),
                ("Mol2Smiles", MolToSmiles()),
                ("ErrorFilter", ErrorFilter()),
            ],
        )

    def test_complex_filter(self) -> None:
        """Test if molecules are filtered correctly by allowed chemical elements."""
        pipeline = ComplexFilterTest._create_pipeline()

        test_params_list_with_results = [
            {
                "params": {},
                "result": [SMILES_BENZENE, SMILES_CHLOROBENZENE],
            },
            {
                "params": {"MultiElementFilter__mode": "all"},
                "result": [],
            },
            {
                "params": {
                    "MultiElementFilter__mode": "any",
                    "MultiElementFilter__pipeline_filter_elements__element_filter_1__add_hydrogens": False,  # noqa: E501
                },
                "result": [SMILES_CHLOROBENZENE],
            },
        ]

        for test_params in test_params_list_with_results:
            pipeline.set_params(**test_params["params"])
            filtered_smiles = pipeline.fit_transform(SMILES_LIST)
            self.assertEqual(filtered_smiles, test_params["result"])

    def test_json_serialization(self) -> None:
        """Test if complex filter can be serialized and deserialized."""
        pipeline = ComplexFilterTest._create_pipeline()
        json_object = recursive_to_json(pipeline)
        newpipeline = recursive_from_json(json_object)
        self.assertEqual(json_object, recursive_to_json(newpipeline))

        pipeline_result = pipeline.fit_transform(SMILES_LIST)
        newpipeline_result = newpipeline.fit_transform(SMILES_LIST)
        self.assertEqual(pipeline_result, newpipeline_result)

    def test_complex_filter_non_unique_names(self) -> None:
        """Test if molecules are filtered correctly by allowed chemical elements."""
        element_filter_1 = ElementFilter({6: 6, 1: 6})
        element_filter_2 = ElementFilter({6: 6, 1: 5, 17: 1})

        with self.assertRaises(ValueError):
            ComplexFilter(
                (("filter_1", element_filter_1), ("filter_1", element_filter_2)),
            )


class SmartsSmilesFilterTest(unittest.TestCase):
    """Unittest for SmartsFilter and SmilesFilter."""

    def test_smarts_smiles_filter(self) -> None:
        """Test if molecules are filtered correctly by allowed SMARTS patterns."""
        smarts_pats: dict[str, IntOrIntCountRange] = {
            "c": (4, None),
            "Cl": 1,
        }
        smarts_filter = SmartsFilter(smarts_pats)

        smiles_pats: dict[str, IntOrIntCountRange] = {
            "c1ccccc1": (1, None),
            "Cl": 1,
        }
        smiles_filter = SmilesFilter(smiles_pats)

        for filter_ in [smarts_filter, smiles_filter]:
            new_input_as_list = list(filter_.filter_elements.keys())
            pipeline = Pipeline(
                [
                    ("Smiles2Mol", SmilesToMol()),
                    ("SmartsFilter", filter_),
                    ("Mol2Smiles", MolToSmiles()),
                    ("ErrorFilter", ErrorFilter()),
                ],
            )

            test_params_list_with_results = [
                {
                    "params": {},
                    "result": [SMILES_BENZENE, SMILES_CHLOROBENZENE, SMILES_CL_BR],
                },
                {
                    "params": {"SmartsFilter__keep_matches": False},
                    "result": [SMILES_ANTIMONY, SMILES_METAL_AU],
                },
                {
                    "params": {
                        "SmartsFilter__mode": "all",
                        "SmartsFilter__keep_matches": True,
                    },
                    "result": [SMILES_CHLOROBENZENE],
                },
                {
                    "params": {
                        "SmartsFilter__keep_matches": True,
                        "SmartsFilter__filter_elements": ["I"],
                    },
                    "result": [],
                },
                {
                    "params": {
                        "SmartsFilter__keep_matches": False,
                        "SmartsFilter__mode": "any",
                        "SmartsFilter__filter_elements": new_input_as_list,
                    },
                    "result": [SMILES_ANTIMONY, SMILES_METAL_AU],
                },
            ]

            for test_params in test_params_list_with_results:
                pipeline.set_params(**test_params["params"])
                filtered_smiles = pipeline.fit_transform(SMILES_LIST)
                self.assertEqual(filtered_smiles, test_params["result"])

    def test_smarts_smiles_filter_wrong_pattern(self) -> None:
        """Test if molecules are filtered correctly by allowed SMARTS patterns."""
        smarts_pats: dict[str, IntOrIntCountRange] = {
            "cIOnk": (4, None),
            "cC": 1,
        }
        with self.assertRaises(ValueError):
            SmartsFilter(smarts_pats)

        smiles_pats: dict[str, IntOrIntCountRange] = {
            "cC": (1, None),
            "Cl": 1,
        }
        with self.assertRaises(ValueError):
            SmilesFilter(smiles_pats)

    def _test_smarts_filter_input_types_helper(
        self,
        smarts_input: Any,
    ) -> None:
        """Test different input types for SmartsFilter.

        Parameters
        ----------
        smarts_input : Any
            The input to the SmartsFilter, can be a list, set, Series or dict of
            patterns.

        """
        pipeline = Pipeline(
            [
                ("Smiles2Mol", SmilesToMol()),
                ("SmartsFilter", SmartsFilter(smarts_input)),
            ],
        )
        filtered_list = pipeline.transform(SMILES_LIST)
        self.assertEqual(len(filtered_list), len(SMILES_LIST))
        self.assertIsInstance(filtered_list[0], InvalidInstance)
        self.assertFalse(isinstance(filtered_list[1], InvalidInstance))
        self.assertFalse(isinstance(filtered_list[2], InvalidInstance))
        self.assertIsInstance(filtered_list[3], InvalidInstance)
        self.assertFalse(isinstance(filtered_list[4], InvalidInstance))

    def test_smarts_filter_input_types(self) -> None:
        """Test different input types for SmartsFilter."""
        self._test_smarts_filter_input_types_helper(["c", "Cl"])
        self._test_smarts_filter_input_types_helper({"c", "Cl"})
        self._test_smarts_filter_input_types_helper(pd.Series(["c", "Cl"]))
        self._test_smarts_filter_input_types_helper({"c": (1, None), "Cl": (1, None)})

    def test_smarts_filter_parallel(self) -> None:
        """Test if molecules are filtered correctly.

        This test runs the SmartsFilter in parallel.

        """
        smarts_pats: dict[str, IntOrIntCountRange] = {
            "c": (4, None),
            "Cl": 1,
            "cc": (1, None),
            "ccc": (1, None),
            "cccc": (1, None),
            "ccccc": (1, None),
            "cccccc": (1, None),
            "c1ccccc1": (1, None),
            "cCl": 1,
        }
        smarts_filter = SmartsFilter(smarts_pats, mode="all", n_jobs=2)
        pipeline = Pipeline(
            [
                ("Smiles2Mol", SmilesToMol()),
                ("SmartsFilter", smarts_filter),
                ("Mol2Smiles", MolToSmiles()),
                ("ErrorFilter", ErrorFilter()),
            ],
        )
        filtered_smiles = pipeline.fit_transform(SMILES_LIST)
        self.assertEqual(filtered_smiles, [SMILES_CHLOROBENZENE])


class RDKitDescriptorsFilterTest(unittest.TestCase):
    """Unittest for RDKitDescriptorsFilter."""

    def test_descriptor_filter(self) -> None:
        """Test if molecules are filtered correctly by allowed descriptors."""
        descriptors: dict[str, FloatCountRange] = {
            "MolWt": (None, 190),
            "NumHAcceptors": (2, 10),
        }

        descriptor_filter = RDKitDescriptorsFilter(descriptors)

        pipeline = Pipeline(
            [
                ("Smiles2Mol", SmilesToMol()),
                ("DescriptorsFilter", descriptor_filter),
                ("Mol2Smiles", MolToSmiles()),
                ("ErrorFilter", ErrorFilter()),
            ],
        )
        test_params_list_with_results = [
            {"params": {}, "result": SMILES_LIST},
            {"params": {"DescriptorsFilter__mode": "all"}, "result": [SMILES_CL_BR]},
            {
                "params": {"DescriptorsFilter__keep_matches": False},
                "result": [
                    SMILES_ANTIMONY,
                    SMILES_BENZENE,
                    SMILES_CHLOROBENZENE,
                    SMILES_METAL_AU,
                ],
            },
            {"params": {"DescriptorsFilter__mode": "any"}, "result": []},
            {
                "params": {
                    "DescriptorsFilter__keep_matches": True,
                    "DescriptorsFilter__filter_elements": {"NumHAcceptors": (2.00, 4)},
                },
                "result": [SMILES_CL_BR],
            },
            {
                "params": {
                    "DescriptorsFilter__filter_elements": {"NumHAcceptors": (1.99, 4)},
                },
                "result": [SMILES_CL_BR],
            },
            {
                "params": {
                    "DescriptorsFilter__filter_elements": {"NumHAcceptors": (2.01, 4)},
                },
                "result": [],
            },
            {
                "params": {
                    "DescriptorsFilter__filter_elements": {"NumHAcceptors": (1, 2.00)},
                },
                "result": [SMILES_CL_BR],
            },
            {
                "params": {
                    "DescriptorsFilter__filter_elements": {"NumHAcceptors": (1, 2.01)},
                },
                "result": [SMILES_CL_BR],
            },
            {
                "params": {
                    "DescriptorsFilter__filter_elements": {"NumHAcceptors": (1, 1.99)},
                },
                "result": [],
            },
        ]

        for test_params in test_params_list_with_results:
            pipeline.set_params(**test_params["params"])
            filtered_smiles = pipeline.fit_transform(SMILES_LIST)
            self.assertEqual(filtered_smiles, test_params["result"])


class MixtureFilterTest(unittest.TestCase):
    """Unittest for MixtureFilter."""

    def test_invalidate_mixtures(self) -> None:
        """Test if mixtures are correctly invalidated."""
        mol_list = ["CCC.CC.C", "c1ccccc1.[Na+].[Cl-]", "c1ccccc1"]
        expected_invalidated_mol_list = [None, None, "c1ccccc1"]

        smi2mol = SmilesToMol()
        mixture_filter = MixtureFilter()
        mol2smi = MolToSmiles()
        error_filter = ErrorFilter.from_element_list([smi2mol, mixture_filter, mol2smi])
        error_replacer = FilterReinserter.from_error_filter(error_filter, None)

        pipeline = Pipeline(
            [
                ("smi2mol", smi2mol),
                ("mixture_filter", mixture_filter),
                ("mol2smi", mol2smi),
                ("error_filter", error_filter),
                ("error_replacer", error_replacer),
            ],
        )
        mols_processed = pipeline.fit_transform(mol_list)
        self.assertEqual(expected_invalidated_mol_list, mols_processed)


class InorganicsFilterTest(unittest.TestCase):
    """Unittest for InorganicsFilter."""

    def test_inorganic_filter(self) -> None:
        """Test if molecules are filtered correctly by allowed chemical elements."""
        smiles2mol = SmilesToMol()
        inorganics_filter = InorganicsFilter()
        mol2smiles = MolToSmiles()
        error_filter = ErrorFilter.from_element_list(
            [smiles2mol, inorganics_filter, mol2smiles],
        )
        pipeline = Pipeline(
            [
                ("Smiles2Mol", smiles2mol),
                ("ElementFilter", inorganics_filter),
                ("Mol2Smiles", mol2smiles),
                ("ErrorFilter", error_filter),
            ],
        )
        filtered_smiles = pipeline.fit_transform(SMILES_LIST)
        self.assertEqual(
            filtered_smiles,
            [SMILES_BENZENE, SMILES_CHLOROBENZENE, SMILES_METAL_AU, SMILES_CL_BR],
        )

        filtered_inroganics = pipeline.fit_transform(["O=C=O", "[O+]#[C-]"])
        self.assertEqual(
            filtered_inroganics,
            [],
        )

        filtered_inroganics = pipeline.fit_transform(InorganicsFilter.CARBON_INORGANICS)
        self.assertEqual(
            filtered_inroganics,
            [],
        )


if __name__ == "__main__":
    unittest.main()



"""Module for checking values."""

from typing import Any

__all__ = ["get_length", "is_empty"]


def is_empty(value: Any) -> bool:
    """Check if value is empty.

    Parameters
    ----------
    value: Any
        Value to be checked.

    Returns
    -------
    bool
        True if value is empty, False otherwise.

    """
    return get_length(value) == 0


def get_length(values: Any) -> int:
    """Get the length of the values as given by the shape or len attribute.

    Parameters
    ----------
    values: Any
        Values to be checked.

    Returns
    -------
    int
        Length of the values.

    Raises
    ------
    TypeError
        If values does not have a shape or len attribute.

    """
    if hasattr(values, "shape"):
        return values.shape[0]
    if hasattr(values, "__len__"):
        return len(values)
    raise TypeError("Values must have a shape or len attribute.")



"""Stratified K-Fold splitter for regression tasks."""

from collections.abc import Iterator

import numpy as np
import pandas as pd
from numpy import typing as npt
from sklearn.model_selection import StratifiedKFold
from typing_extensions import override


class PercentileStratifiedKFold(StratifiedKFold):  # pylint: disable=abstract-method
    """Stratified K-Fold splitter for regression tasks using percentile-based binning.

    This splitter bins continuous target values into percentiles, which by definition
    are equally sized groups, and then applies stratified sampling to ensure that each
    fold has a balanced distribution of target values across the percentiles.

    """

    def __init__(
        self,
        n_splits: int = 5,
        *,
        n_groups: int = 10,
        shuffle: bool = True,
        random_state: int | None = None,
    ) -> None:
        """Initialize the PercentileStratifiedRegressionKFold.

        Parameters
        ----------
        n_splits : int, default=5
            Number of folds for cross-validation.
        n_groups : int, default=10
            Number of percentile groups to create for stratification.
        shuffle : bool, default=True
            Whether to shuffle the data before splitting into batches.
        random_state : int | None, optional
            Random state for reproducibility.

        """
        self.n_groups = n_groups
        super().__init__(n_splits=n_splits, shuffle=shuffle, random_state=random_state)

    @override
    def split(
        self,
        X: npt.ArrayLike,
        y: npt.ArrayLike,
        groups: npt.ArrayLike | None = None,
    ) -> Iterator[tuple[npt.NDArray[np.int_], npt.NDArray[np.int_]]]:
        """Generate indices to split data into training and test set.

        Parameters
        ----------
        X : npt.ArrayLike
            The model input data.
        y : npt.ArrayLike
            The target data.
        groups : npt.ArrayLike | None, optional
            The group data, by default None.

        Yields
        ------
        npt.NDArray[np.int_]
            The training indices.
        npt.NDArray[np.int_]
            The test indices.

        Raises
        ------
        ValueError
            If n_groups is greater than the number of unique target values.

        """
        n_groups = self.n_groups
        y = np.asarray(y, dtype=np.float64)
        if self.n_groups > len(np.unique(y)):
            raise ValueError(
                f"n_groups ({self.n_groups}) is greater than the number of unique "
                f"target values ({len(np.unique(y))})!",
            )

        try:
            y_binned = pd.qcut(y, n_groups, labels=False, duplicates="raise")
        except ValueError as e:
            raise ValueError(
                "Too many identical values in y for the specified number of groups!",
            ) from e

        yield from super().split(X, y_binned)



"""Conformal prediction wrappers for classification and regression using crepes.

This module provides a unified interface for conformal prediction with 4 main classes:
- ConformalClassifier: Single-model conformal classification
- CrossConformalClassifier: Cross-validation conformal classification
- ConformalRegressor: Single-model conformal regression
- CrossConformalRegressor: Cross-validation conformal regression

All classes use composition with crepes and provide full sklearn compatibility.
"""

# Import the four main conformal prediction classes via the legacy module path.
from molpipeline.experimental.uncertainty.conformal import (
    ConformalClassifier,
    ConformalRegressor,
    CrossConformalClassifier,
    CrossConformalRegressor,
    _apply_antitonic_regressors,
    _fit_antitonic_regressors,
)

# Import nonconformity functions via the legacy utils module path.
from molpipeline.experimental.uncertainty.utils import (
    HingeNonconformity,
    LogNonconformity,
    MarginNonconformity,
    NonconformityFunctor,
    SVMMarginNonconformity,
    create_nonconformity_function,
)

# Export all the important classes
__all__ = [
    "ConformalClassifier",
    "ConformalRegressor",
    "CrossConformalClassifier",
    "CrossConformalRegressor",
    "HingeNonconformity",
    "LogNonconformity",
    "MarginNonconformity",
    "NonconformityFunctor",
    "SVMMarginNonconformity",
    "_apply_antitonic_regressors",
    "_fit_antitonic_regressors",
    "create_nonconformity_function",
]



"""Initialize module for data splitting."""

from molpipeline.experimental.model_selection.splitter.bootstrap_splitter import (
    BootstrapSplit,
)
from molpipeline.experimental.model_selection.splitter.data_repetition_splitter import (
    DataRepetitionSplit,
)
from molpipeline.experimental.model_selection.splitter.group_addition_splitter import (
    GroupAdditionSplit,
)
from molpipeline.experimental.model_selection.splitter.group_shuffle_splitter import (
    GroupShuffleSplit,
    SplitModeOption,
)
from molpipeline.experimental.model_selection.splitter.stratified_regression import (
    PercentileStratifiedKFold,
)
from molpipeline.experimental.model_selection.splitter.time_threshold_splitter import (
    TimeThresholdSplitter,
)

__all__ = [
    "BootstrapSplit",
    "DataRepetitionSplit",
    "GroupAdditionSplit",
    "GroupShuffleSplit",
    "PercentileStratifiedKFold",
    "SplitModeOption",
    "TimeThresholdSplitter",
]



"""Tests for URL file loading utilities."""

import unittest
from unittest import mock

from sklearn.base import clone

from molpipeline.utils.file_loading.url_file_loading import URLFileLoader
from molpipeline.utils.json_operations import recursive_from_json, recursive_to_json


class TestURLFileLoader(unittest.TestCase):
    """Unit tests for the URLFileLoader class."""

    def setUp(self) -> None:
        """Set up test variables."""
        self.test_url = (
            "https://github.com/basf/MolPipeline/blob/main/.github/molpipeline.png"
        )
        self.loader = URLFileLoader(url=self.test_url, timeout=10)

    def test_initialization(self) -> None:
        """Test initialization of URLFileLoader."""
        self.assertEqual(self.loader.url, self.test_url)
        self.assertEqual(self.loader.timeout, 10)

    def test_get_params(self) -> None:
        """Test get_params method."""
        params = self.loader.get_params()
        self.assertEqual(params, {"url": self.test_url, "timeout": 10})

    def test_set_params(self) -> None:
        """Test set_params method."""
        new_url = "https://www.test.com"
        new_timeout = 20
        self.loader.set_params(url=new_url, timeout=new_timeout)
        self.assertEqual(self.loader.url, new_url)
        self.assertEqual(self.loader.timeout, new_timeout)

    def test_load_file(self) -> None:
        """Test load_file method."""
        mock_message = "Mocked content"
        with mock.patch("requests.get") as mock_get:
            mock_response = mock.MagicMock()
            mock_response.content = mock_message.encode("utf-8")
            mock_response.raise_for_status = mock.MagicMock()
            mock_get.return_value = mock_response
            loaded_file = self.loader.load_file()
            self.assertEqual(mock_get.call_count, 1)

        self.assertEqual(loaded_file.decode(), mock_message)

    def test_serialization(self) -> None:
        """Test serialization and deserialization of URLFileLoader."""
        json_data = recursive_to_json(self.loader)
        new_loader = recursive_from_json(json_data)
        self.assertIsInstance(new_loader, URLFileLoader)
        self.assertEqual(new_loader.url, self.loader.url)
        self.assertEqual(new_loader.timeout, self.loader.timeout)

    def test_cloning(self) -> None:
        """Test cloning of URLFileLoader."""
        cloned_loader: URLFileLoader = clone(self.loader)  # type: ignore
        self.assertNotEqual(cloned_loader, self.loader)
        self.assertIsInstance(cloned_loader, URLFileLoader)
        self.assertEqual(cloned_loader.url, self.loader.url)
        self.assertEqual(cloned_loader.timeout, self.loader.timeout)



"""Defines a pipeline is exposed to the user, accessible via pipeline."""

# pylint: disable=too-many-lines

from collections.abc import Iterable
from copy import deepcopy
from typing import Any, Literal, Self

import joblib
import numpy as np
import numpy.typing as npt
from loguru import logger
from sklearn.base import _fit_context, clone
from sklearn.pipeline import Pipeline as _Pipeline
from sklearn.pipeline import _final_estimator_has, _fit_transform_one
from sklearn.utils import Bunch
from sklearn.utils._tags import Tags, get_tags
from sklearn.utils.metadata_routing import (
    MetadataRouter,
    MethodMapping,
    _routing_enabled,
    process_routing,
)
from sklearn.utils.metaestimators import available_if
from sklearn.utils.validation import check_memory
from typing_extensions import override

from molpipeline.abstract_pipeline_elements.core import ABCPipelineElement
from molpipeline.error_handling import ErrorFilter, FilterReinserter
from molpipeline.pipeline._molpipeline import _MolPipeline
from molpipeline.post_prediction import (
    PostPredictionTransformation,
    PostPredictionWrapper,
)
from molpipeline.utils.logging import print_elapsed_time
from molpipeline.utils.molpipeline_types import (
    AnyElement,
    AnyPredictor,
    AnyStep,
    AnyTransformer,
)
from molpipeline.utils.value_checks import is_empty

__all__ = ["Pipeline"]

# Type definitions
# Cannot be moved to utils.molpipeline_types due to circular imports


_IndexedStep = tuple[int, str, AnyElement]
_AggStep = tuple[list[int], list[str], _MolPipeline]
_AggregatedPipelineStep = _IndexedStep | _AggStep


class Pipeline(_Pipeline):
    """Defines the pipeline which handles pipeline elements."""

    steps: list[AnyStep]
    #  * Adapted methods from sklearn.pipeline.Pipeline *

    def __init__(
        self,
        steps: list[AnyStep],
        *,
        memory: str | joblib.Memory | None = None,
        verbose: bool = False,
        n_jobs: int = 1,
    ) -> None:
        """Initialize Pipeline.

        Parameters
        ----------
        steps: list[tuple[str, Union[AnyTransformer, AnyPredictor, ABCPipelineElement]]]
            List of (name, Estimator) tuples.
        memory: str | joblib.Memory | None, optional
            Path to cache transformers.
        verbose: bool, optional
            If True, print additional information.
        n_jobs: int, optional
            Number of cores used for aggregated steps.

        """
        super().__init__(steps, memory=memory, verbose=verbose)
        self.n_jobs = n_jobs
        self._set_error_resinserter()

    def _set_error_resinserter(self) -> None:
        """Connect the error resinserters with the error filters."""
        error_replacer_list = [
            e_filler
            for _, e_filler in self.steps
            if isinstance(e_filler, FilterReinserter)
        ]
        error_filter_list = [
            n_filter for _, n_filter in self.steps if isinstance(n_filter, ErrorFilter)
        ]
        for step in self.steps:
            if isinstance(step[1], PostPredictionWrapper) and isinstance(
                step[1].wrapped_estimator,
                FilterReinserter,
            ):
                error_replacer_list.append(step[1].wrapped_estimator)
        for error_replacer in error_replacer_list:
            error_replacer.select_error_filter(error_filter_list)

    def _validate_steps(self) -> None:
        """Validate the steps.

        Raises
        ------
        TypeError
            If the steps do not implement fit and transform or are not 'passthrough'.

        """
        names = [name for name, _ in self.steps]

        # validate names
        self._validate_names(names)

        # validate estimators
        non_post_processing_steps = [e for _, _, e in self._agg_non_postpred_steps()]
        transformer_list = non_post_processing_steps[:-1]
        estimator = non_post_processing_steps[-1]

        for transformer in transformer_list:
            if transformer is None or transformer == "passthrough":
                continue
            if not (
                hasattr(transformer, "fit") or hasattr(transformer, "fit_transform")
            ) or not hasattr(transformer, "transform"):
                raise TypeError(
                    f"All intermediate steps should be "
                    f"transformers and implement fit and transform "
                    f"or be the string 'passthrough' "
                    f"'{transformer}' (type {type(transformer)}) doesn't",
                )

        # We allow last estimator to be None as an identity transformation
        if (
            estimator is not None
            and estimator != "passthrough"
            and not hasattr(estimator, "fit")
        ):
            raise TypeError(
                f"Last step of Pipeline should implement fit "
                f"or be the string 'passthrough'. "
                f"'{estimator}' (type {type(estimator)}) doesn't",
            )

        # validate post-processing steps
        # Calling steps automatically validates them
        _ = self._post_processing_steps()

    def _iter(
        self,
        with_final: bool = True,
        filter_passthrough: bool = True,
    ) -> Iterable[_AggregatedPipelineStep]:
        """Iterate over all non post-processing steps.

        Steps where children of a ABCPipelineElement are aggregated to a MolPipeline.

        Parameters
        ----------
        with_final: bool, optional
            If True, the final estimator is included.
        filter_passthrough: bool, optional
            If True, passthrough steps are filtered out.

        Yields
        ------
        _AggregatedPipelineStep
            The _AggregatedPipelineStep is composed of the index, the name and the
            transformer.

        Raises
        ------
        AssertionError
            If the pipeline has no steps.

        """
        last_element: _AggregatedPipelineStep | None = None

        # This loop delays the output by one in order to identify the last step
        for step in self._agg_non_postpred_steps():
            # Only happens for the first step
            if last_element is None:
                last_element = step
                continue
            if not filter_passthrough or (
                step[2] is not None and step[2] != "passthrough"
            ):
                yield last_element
            last_element = step

        # This can only happen if no steps are set.
        if last_element is None:
            raise AssertionError("Pipeline needs to have at least one step!")

        if (
            with_final
            and last_element[2] is not None
            and last_element[2] != "passthrough"
        ):
            yield last_element

    @property
    def _estimator_type(self) -> Any:
        """Return the estimator type."""
        if self._final_estimator is None or self._final_estimator == "passthrough":
            return None
        if hasattr(self._final_estimator, "_estimator_type"):
            return self._final_estimator._estimator_type  # noqa: SLF001
        return None

    @property
    def _final_estimator(
        self,
    ) -> (
        Literal["passthrough"]
        | AnyTransformer
        | AnyPredictor
        | _MolPipeline
        | ABCPipelineElement
        | FilterReinserter[Any]
    ):
        """Return the lst estimator which is not a PostprocessingTransformer."""
        element_list = list(self._agg_non_postpred_steps())
        last_element = element_list[-1]
        return last_element[2]

    # pylint: disable=too-many-locals,too-many-branches
    @override
    def _fit(
        self,
        X: Any,
        y: Any = None,
        routed_params: dict[str, Any] | None = None,
        raw_params: dict[str, Any] | None = None,
    ) -> tuple[Any, Any]:
        """Fit the model by fitting all transformers except the final estimator.

        Data can be subsetted by the transformers.

        Parameters
        ----------
        X : Any
            Training data.
        y : Any, optional (default=None)
            Training objectives.
        routed_params : dict[str, Any], optional
            Parameters for each step as returned by process_routing.
            Although this is marked as optional, it should not be None.
            The awkward (argward?) typing is due to inheritance from sklearn.
            Can be an empty dictionary.
        raw_params : dict[str, Any], optional
            Parameters passed by the user, used when `transform_input`

        Returns
        -------
        tuple[Any, Any]
            The transformed data and the transformed objectives.

        Raises
        ------
        AssertionError
            If routed_params is None or if the transformer is 'passthrough'.
        AssertionError
            If the names are a list and the step is not a Pipeline.

        """
        # shallow copy of steps - this should really be steps_
        self.steps = list(self.steps)
        self._validate_steps()
        if routed_params is None:
            raise AssertionError("routed_params should not be None.")

        # Set up the memory
        memory: joblib.Memory = check_memory(self.memory)

        fit_transform_one_cached = memory.cache(_fit_transform_one)
        for step in self._iter(with_final=False, filter_passthrough=False):
            step_idx, name, transformer = step
            if transformer is None or transformer == "passthrough":
                with print_elapsed_time("Pipeline", self._log_message(step_idx)):
                    continue

            if hasattr(memory, "location") and memory.location is None:
                # we do not clone when caching is disabled to
                # preserve backward compatibility
                cloned_transformer = transformer
            else:
                cloned_transformer = clone(transformer)
            if isinstance(cloned_transformer, _MolPipeline):
                if routed_params:
                    step_params = {
                        "element_parameters": [routed_params[n] for n in name],
                    }
                else:
                    step_params = {}
            elif isinstance(name, list):
                raise AssertionError(
                    "Names should not be a list, when the step is not a Pipeline",
                )
            else:
                step_params = self._get_metadata_for_step(
                    step_idx=step_idx,
                    step_params=routed_params[name],
                    all_params=raw_params,
                )

            # Fit or load from cache the current transformer
            X, fitted_transformer = fit_transform_one_cached(  # noqa: N806
                cloned_transformer,
                X,
                y,
                None,
                message_clsname="Pipeline",
                message=self._log_message(step_idx),
                params=step_params,
            )
            # Replace the transformer of the step with the fitted
            # transformer. This is necessary when loading the transformer
            # from the cache.
            if isinstance(fitted_transformer, _MolPipeline):
                ele_list = fitted_transformer.element_list
                if not isinstance(name, list) or not isinstance(step_idx, list):
                    raise AssertionError()
                if not len(name) == len(step_idx) == len(ele_list):
                    raise AssertionError()
                for idx_i, name_i, ele_i in zip(step_idx, name, ele_list, strict=True):
                    self.steps[idx_i] = (name_i, ele_i)
                if y is not None:
                    y = fitted_transformer.co_transform(y)
                for idx_i, name_i, ele_i in zip(step_idx, name, ele_list, strict=True):
                    self.steps[idx_i] = (name_i, ele_i)
                self._set_error_resinserter()
            elif isinstance(name, list) or isinstance(step_idx, list):
                raise AssertionError()
            else:
                self.steps[step_idx] = (name, fitted_transformer)
            if is_empty(X):
                return np.array([]), np.array([])
        return X, y

    def _transform(
        self,
        X: Any,  # noqa: N803
        routed_params: Bunch,
    ) -> Any:
        """Transform the data, and skip final estimator.

        Call `transform` of each transformer in the pipeline except the last one,

        Parameters
        ----------
        X : iterable
            Data to predict on. Must fulfill input requirements of first step
            of the pipeline.
        routed_params: Bunch
            Parameters for each step as returned by process_routing

        Returns
        -------
        Any
            Result of calling `transform` on the second last estimator.

        Raises
        ------
        AssertionError
            If one of the transformers is 'passthrough' or does not implement
            `transform`.

        """
        iter_input = X
        do_routing = _routing_enabled()
        if do_routing:
            logger.warning("Routing is enabled and NOT fully tested!")

        for _, name, transform in self._iter(with_final=False):
            if is_empty(iter_input):
                if isinstance(transform, _MolPipeline):
                    _ = transform.transform(iter_input)
                iter_input = []
                break
            if transform == "passthrough":
                raise AssertionError("Passthrough should have been filtered out.")
            if hasattr(transform, "transform"):
                if do_routing:
                    iter_input = transform.transform(  # type: ignore[call-arg]
                        iter_input,
                        routed_params[name].transform,
                    )
                else:
                    iter_input = transform.transform(iter_input)
            else:
                raise AssertionError(
                    f"Non transformer ocurred in transformation step: {transform}.",
                )
        return iter_input

    # * New implemented methods *
    def _non_post_processing_steps(
        self,
    ) -> list[AnyStep]:
        """Return all steps before the first PostPredictionTransformation.

        Returns
        -------
        list[AnyStep]
            List of steps before the first PostPredictionTransformation.

        Raises
        ------
        AssertionError
            If a PostPredictionTransformation is found before the last step.

        """
        non_post_processing_steps: list[AnyStep] = []
        start_adding = False
        for step_name, step_estimator in self.steps[::-1]:
            if not isinstance(step_estimator, PostPredictionTransformation):
                start_adding = True
            if start_adding:
                if isinstance(step_estimator, PostPredictionTransformation):
                    raise AssertionError(
                        "PipelineElement of type PostPredictionTransformation occured "
                        "before the last step.",
                    )
                non_post_processing_steps.append((step_name, step_estimator))
        return list(non_post_processing_steps[::-1])

    def _post_processing_steps(self) -> list[tuple[str, PostPredictionTransformation]]:
        """Return last steps which are PostPredictionTransformation.

        Returns
        -------
        list[tuple[str, PostPredictionTransformation]]
            List of tuples containing the name and the PostPredictionTransformation.

        """
        post_processing_steps = []
        for step_name, step_estimator in self.steps[::-1]:
            if isinstance(step_estimator, PostPredictionTransformation):
                post_processing_steps.append((step_name, step_estimator))
            else:
                break
        return list(post_processing_steps[::-1])

    def _agg_non_postpred_steps(
        self,
    ) -> Iterable[_AggregatedPipelineStep]:
        """Generate (idx, (name, trans)) tuples from self.steps.

        When filter_passthrough is True, 'passthrough' and None transformers
        are filtered out.

        Yields
        ------
        _AggregatedPipelineStep
            The _AggregatedPipelineStep is composed of the index, the name and the
            transformer.

        """
        aggregated_transformer_list = []
        for i, (name_i, step_i) in enumerate(self._non_post_processing_steps()):
            if isinstance(step_i, ABCPipelineElement):
                aggregated_transformer_list.append((i, name_i, step_i))
            else:
                if aggregated_transformer_list:
                    index_list = [step[0] for step in aggregated_transformer_list]
                    name_list = [step[1] for step in aggregated_transformer_list]
                    transformer_list = [step[2] for step in aggregated_transformer_list]
                    if len(aggregated_transformer_list) == 1:
                        yield index_list[0], name_list[0], transformer_list[0]
                    else:
                        pipeline = _MolPipeline(transformer_list, n_jobs=self.n_jobs)
                        yield index_list, name_list, pipeline
                    aggregated_transformer_list = []
                yield i, name_i, step_i

        # yield last step if anything remains
        if aggregated_transformer_list:
            index_list = [step[0] for step in aggregated_transformer_list]
            name_list = [step[1] for step in aggregated_transformer_list]
            transformer_list = [step[2] for step in aggregated_transformer_list]

            if len(aggregated_transformer_list) == 1:
                yield index_list[0], name_list[0], transformer_list[0]

            elif len(aggregated_transformer_list) > 1:
                pipeline = _MolPipeline(transformer_list, n_jobs=self.n_jobs)
                yield index_list, name_list, pipeline

    @_fit_context(
        # estimators in Pipeline.steps are not validated yet
        prefer_skip_nested_validation=False,
    )
    @override
    def fit(self, X: Any, y: Any = None, **fit_params: Any) -> Self:
        """Fit the model.

        Fit all the transformers one after the other and transform the
        data. Finally, fit the transformed data using the final estimator.

        Parameters
        ----------
        X : iterable
            Training data. Must fulfill input requirements of first step of the
            pipeline.

        y : iterable, default=None
            Training targets. Must fulfill label requirements for all steps of
            the pipeline.

        **fit_params : dict of string -> object
            Parameters passed to the ``fit`` method of each step, where
            each parameter name is prefixed such that parameter ``p`` for step
            ``s`` has key ``s__p``.

        Returns
        -------
        self : object
            Pipeline with fitted steps.

        """
        routed_params = self._check_method_params(method="fit", props=fit_params)
        Xt, yt = self._fit(X, y, routed_params)  # noqa: N806
        with print_elapsed_time("Pipeline", self._log_message(len(self.steps) - 1)):
            if self._final_estimator != "passthrough":
                if is_empty(Xt):
                    logger.warning(
                        "All input rows were filtered out! Model is not fitted!",
                    )
                else:
                    fit_params_last_step = routed_params[
                        self._non_post_processing_steps()[-1][0]
                    ]
                    self._final_estimator.fit(Xt, yt, **fit_params_last_step["fit"])

        return self

    @override
    def _can_fit_transform(self) -> bool:
        """Check if the final estimator can fit_transform or is passthrough.

        Returns
        -------
        bool
            True if the final estimator can fit_transform or is passthrough.

        """
        return (
            self._final_estimator == "passthrough"
            or hasattr(self._final_estimator, "transform")
            or hasattr(self._final_estimator, "fit_transform")
        )

    def _can_decision_function(self) -> bool:
        """Check if the final estimator implements decision_function.

        Returns
        -------
        bool
            True if the final estimator implements decision_function.

        """
        return hasattr(self._final_estimator, "decision_function")

    @available_if(_can_fit_transform)
    @_fit_context(
        # estimators in Pipeline.steps are not validated yet
        prefer_skip_nested_validation=False,
    )
    @override
    def fit_transform(self, X: Any, y: Any = None, **params: Any) -> Any:
        """Fit the model and transform with the final estimator.

        Fits all the transformers one after the other and transform the
        data. Then uses `fit_transform` on transformed data with the final
        estimator.

        Parameters
        ----------
        X : iterable
            Training data. Must fulfill input requirements of first step of the
            pipeline.

        y : iterable, default=None
            Training targets. Must fulfill label requirements for all steps of
            the pipeline.

        **params : Any
            Parameters passed to the ``fit`` method of each step, where
            each parameter name is prefixed such that parameter ``p`` for step
            ``s`` has key ``s__p``.

        Returns
        -------
        Xt : ndarray of shape (n_samples, n_transformed_features)
            Transformed samples.

        Raises
        ------
        TypeError
            If the last step does not implement `fit_transform` or `fit` and
            `transform`.

        """
        routed_params = self._check_method_params(method="fit_transform", props=params)
        iter_input, iter_label = self._fit(X, y, routed_params)
        last_step = self._final_estimator
        with print_elapsed_time("Pipeline", self._log_message(len(self.steps) - 1)):
            if last_step == "passthrough":
                pass
            elif is_empty(iter_input):
                logger.warning("All input rows were filtered out! Model is not fitted!")
            else:
                last_step_params = routed_params[
                    self._non_post_processing_steps()[-1][0]
                ]
                if hasattr(last_step, "fit_transform"):
                    iter_input = last_step.fit_transform(
                        iter_input,
                        iter_label,
                        **last_step_params["fit_transform"],
                    )
                elif hasattr(last_step, "transform") and hasattr(last_step, "fit"):
                    last_step.fit(iter_input, iter_label, **last_step_params["fit"])
                    iter_input = last_step.transform(
                        iter_input,
                        **last_step_params["transform"],
                    )
                else:
                    raise TypeError(
                        f"fit_transform of the final estimator"
                        f" {last_step.__class__.__name__} {last_step_params} does not "
                        f"match fit_transform of Pipeline {self.__class__.__name__}",
                    )
            for _, post_element in self._post_processing_steps():
                iter_input = post_element.fit_transform(iter_input, iter_label)
        return iter_input

    @available_if(_final_estimator_has("predict"))
    @override
    def predict(self, X: Any, **params: Any) -> Any:
        """Transform the data, and apply `predict` with the final estimator.

        Call `transform` of each transformer in the pipeline. The transformed
        data are finally passed to the final estimator that calls `predict`
        method. Only valid if the final estimator implements `predict`.

        Parameters
        ----------
        X : iterable
            Data to predict on. Must fulfill input requirements of first step
            of the pipeline.

        **params : dict of string -> object
            Parameters to the `predict` called at the end of all
            transformations in the pipeline. Note that while this may be
            used to return uncertainties from some models with return_std
            or return_cov, uncertainties that are generated by the
            transformations in the pipeline are not propagated to the
            final estimator.

        Returns
        -------
        y_pred : ndarray
            Result of calling `predict` on the final estimator.

        Raises
        ------
        AssertionError
            If the final estimator does not implement `predict`.
            In this case this function should not be available.

        """
        if _routing_enabled():
            routed_params = process_routing(self, "predict", **params)
        else:
            routed_params = process_routing(self, "predict")

        iter_input = self._transform(X, routed_params)

        if self._final_estimator == "passthrough":
            pass
        elif is_empty(iter_input):
            iter_input = []
        elif hasattr(self._final_estimator, "predict"):
            if _routing_enabled():
                iter_input = self._final_estimator.predict(
                    iter_input,
                    **routed_params[self._non_post_processing_steps()[-1][0]].predict,
                )
            else:
                iter_input = self._final_estimator.predict(iter_input, **params)
        else:
            raise AssertionError(
                "Final estimator does not implement predict, hence this function "
                "should not be available.",
            )
        for _, post_element in self._post_processing_steps():
            iter_input = post_element.transform(iter_input)
        return iter_input

    @available_if(_final_estimator_has("fit_predict"))
    @_fit_context(
        # estimators in Pipeline.steps are not validated yet
        prefer_skip_nested_validation=False,
    )
    @override
    def fit_predict(self, X: Any, y: Any = None, **params: Any) -> Any:
        """Transform the data, and apply `fit_predict` with the final estimator.

        Call `fit_transform` of each transformer in the pipeline. The
        transformed data are finally passed to the final estimator that calls
        `fit_predict` method. Only valid if the final estimator implements
        `fit_predict`.

        Parameters
        ----------
        X : iterable
            Training data. Must fulfill input requirements of first step of
            the pipeline.

        y : iterable, default=None
            Training targets. Must fulfill label requirements for all steps
            of the pipeline.

        **params : dict of string -> object
            Parameters passed to the ``fit`` method of each step, where
            each parameter name is prefixed such that parameter ``p`` for step
            ``s`` has key ``s__p``.

        Returns
        -------
        y_pred : ndarray
            Result of calling `fit_predict` on the final estimator.

        Raises
        ------
        AssertionError
            If the final estimator does not implement `fit_predict`.
            In this case this function should not be available.

        """
        routed_params = self._check_method_params(method="fit_predict", props=params)
        iter_input, iter_label = self._fit(X, y, routed_params)

        params_last_step = routed_params[self._non_post_processing_steps()[-1][0]]
        with print_elapsed_time("Pipeline", self._log_message(len(self.steps) - 1)):
            if self._final_estimator == "passthrough":
                y_pred = iter_input
            elif is_empty(iter_input):
                logger.warning("All input rows were filtered out! Model is not fitted!")
                iter_input = []
                y_pred = []
            elif hasattr(self._final_estimator, "fit_predict"):
                y_pred = self._final_estimator.fit_predict(
                    iter_input,
                    iter_label,
                    **params_last_step.get("fit_predict", {}),
                )
            else:
                raise AssertionError(
                    "Final estimator does not implement fit_predict, "
                    "hence this function should not be available.",
                )
            for _, post_element in self._post_processing_steps():
                y_pred = post_element.fit_transform(y_pred, iter_label)
        return y_pred

    @available_if(_final_estimator_has("predict_proba"))
    @override
    def predict_proba(self, X: Any, **params: Any) -> Any:
        """Transform the data, and apply `predict_proba` with the final estimator.

        Call `transform` of each transformer in the pipeline. The transformed
        data are finally passed to the final estimator that calls `predict_proba`
        method. Only valid if the final estimator implements `predict_proba`.

        Parameters
        ----------
        X : iterable
            Data to predict on. Must fulfill input requirements of first step
            of the pipeline.

        **params : dict of string -> object
            Parameters to the ``predict`` called at the end of all
            transformations in the pipeline. Note that while this may be
            used to return uncertainties from some models with return_std
            or return_cov, uncertainties that are generated by the
            transformations in the pipeline are not propagated to the
            final estimator.

        Returns
        -------
        y_pred : ndarray
            Result of calling `predict_proba` on the final estimator.

        Raises
        ------
        AssertionError
            If the final estimator does not implement `predict_proba`.
            In this case this function should not be available.

        """
        routed_params = process_routing(self, "predict_proba", **params)
        iter_input = self._transform(X, routed_params)

        if self._final_estimator == "passthrough":
            pass
        elif is_empty(iter_input):
            iter_input = []
        elif hasattr(self._final_estimator, "predict_proba"):
            if _routing_enabled():
                iter_input = self._final_estimator.predict_proba(
                    iter_input,
                    **routed_params[
                        self._non_post_processing_steps()[-1][0]
                    ].predict_proba,
                )
            else:
                iter_input = self._final_estimator.predict_proba(iter_input, **params)
        else:
            raise AssertionError(
                "Final estimator does not implement predict_proba, "
                "hence this function should not be available.",
            )
        for _, post_element in self._post_processing_steps():
            iter_input = post_element.transform(iter_input)
        return iter_input

    @override
    def _can_transform(self) -> bool:
        """Check if the final estimator can transform or is passthrough.

        Returns
        -------
        bool
            True if the final estimator can transform or is passthrough.

        """
        return self._final_estimator == "passthrough" or hasattr(
            self._final_estimator,
            "transform",
        )

    @available_if(_can_transform)
    @override
    def transform(self, X: Any, **params: Any) -> Any:
        """Transform the data, and apply `transform` with the final estimator.

        Call `transform` of each transformer in the pipeline. The transformed
        data are finally passed to the final estimator that calls
        `transform` method. Only valid if the final estimator
        implements `transform`.

        This also works where final estimator is `None` in which case all prior
        transformations are applied.

        Parameters
        ----------
        X : iterable
            Data to transform. Must fulfill input requirements of first step
            of the pipeline.
        **params : Any
            Parameters to the ``transform`` method of each estimator.

        Returns
        -------
        Xt : ndarray of shape (n_samples, n_transformed_features)
            Transformed data.

        Raises
        ------
        AssertionError
            If the final estimator does not implement `transform` or
            `fit_transform` or is passthrough.

        """
        routed_params = process_routing(self, "transform", **params)
        iter_input = X
        for _, name, transform in self._iter():
            if transform == "passthrough":
                continue
            if is_empty(iter_input):
                # This is done to prime the error filters
                if isinstance(transform, _MolPipeline):
                    _ = transform.transform(iter_input)
                iter_input = []
                break
            if hasattr(transform, "transform"):
                iter_input = transform.transform(
                    iter_input,
                    **routed_params[name].transform,
                )
            else:
                raise AssertionError(
                    "Non transformer ocurred in transformation step."
                    "This should have been caught in the validation step.",
                )
        for _, post_element in self._post_processing_steps():
            iter_input = post_element.transform(iter_input, **params)
        return iter_input

    @available_if(_can_decision_function)
    @override
    def decision_function(self, X: Any, **params: Any) -> Any:
        """Transform the data, and apply `decision_function` with the final estimator.

        Parameters
        ----------
        X : iterable
            Data to transform. Must fulfill input requirements of first step
            of the pipeline.
        **params : Any
            Parameters to the ``decision_function`` method of the final estimator.

        Returns
        -------
        Any
            Result of calling `decision_function` on the final estimator.

        Raises
        ------
        AssertionError
            If the final estimator does not implement `decision_function`.

        """
        if _routing_enabled():
            routed_params = process_routing(self, "decision_function", **params)
        else:
            routed_params = process_routing(self, "decision_function")

        iter_input = self._transform(X, routed_params)
        if self._final_estimator == "passthrough":
            pass
        elif is_empty(iter_input):
            iter_input = []
        elif hasattr(self._final_estimator, "decision_function"):
            if _routing_enabled():
                iter_input = self._final_estimator.decision_function(
                    iter_input,
                    **routed_params[self._final_estimator].predict,
                )
            else:
                iter_input = self._final_estimator.decision_function(
                    iter_input,
                    **params,
                )
        else:
            raise AssertionError(
                "Final estimator does not implement `decision_function`, "
                "hence this function should not be available.",
            )
        for _, post_element in self._post_processing_steps():
            iter_input = post_element.transform(iter_input)
        return iter_input

    @property
    def classes_(self) -> list[Any] | npt.NDArray[Any]:
        """Return the classes of the last element.

        PostPredictionTransformation elements are not considered as last element.

        Raises
        ------
        ValueError
            If the last step is passthrough or has no classes_ attribute.

        """
        check_last = [
            step
            for step in self.steps
            if not isinstance(step[1], PostPredictionTransformation)
        ]
        last_step = check_last[-1][1]
        if last_step == "passthrough":
            raise ValueError("Last step is passthrough.")
        if hasattr(last_step, "classes_"):
            return last_step.classes_
        raise ValueError("Last step has no classes_ attribute.")

    def __sklearn_tags__(self) -> Tags:
        """Return the sklearn tags.

        Returns
        -------
        Tags
            The sklearn tags.

        Notes
        -----
        This method is copied from the original sklearn implementation.
        Changes are marked with a comment.

        """
        tags = super().__sklearn_tags__()

        if not self.steps:
            return tags

        try:
            if self.steps[0][1] is not None and self.steps[0][1] != "passthrough":
                tags.input_tags.pairwise = get_tags(
                    self.steps[0][1],
                ).input_tags.pairwise
            # WARNING: the sparse tag can be incorrect.
            # Some Pipelines accepting sparse data are wrongly tagged sparse=False.
            # For example Pipeline([PCA(), estimator]) accepts sparse data
            # even if the estimator doesn't as PCA outputs a dense array.
            tags.input_tags.sparse = all(
                get_tags(step).input_tags.sparse
                for name, step in self.steps
                if step != "passthrough"
            )
        except (ValueError, AttributeError, TypeError):
            # This happens when the `steps` is not a list of (name, estimator)
            # tuples and `fit` is not called yet to validate the steps.
            pass

        try:
            # Only the _final_estimator is changed from the original implementation is
            # changed in the following 2 lines
            if (
                self._final_estimator is not None
                and self._final_estimator != "passthrough"
            ):
                last_step_tags = get_tags(self._final_estimator)
                tags.estimator_type = last_step_tags.estimator_type
                tags.target_tags.multi_output = last_step_tags.target_tags.multi_output
                tags.classifier_tags = deepcopy(last_step_tags.classifier_tags)
                tags.regressor_tags = deepcopy(last_step_tags.regressor_tags)
                tags.transformer_tags = deepcopy(last_step_tags.transformer_tags)
        except (ValueError, AttributeError, TypeError):
            # This happens when the `steps` is not a list of (name, estimator)
            # tuples and `fit` is not called yet to validate the steps.
            pass

        return tags

    def get_metadata_routing(self) -> MetadataRouter:
        """Get metadata routing of this object.

        Please check :ref:`User Guide <metadata_routing>` on how the routing
        mechanism works.

        Returns
        -------
        MetadataRouter
            A :class:`~sklearn.utils.metadata_routing.MetadataRouter` encapsulating
            routing information.

        Notes
        -----
        This method is copied from the original sklearn implementation.
        Changes are marked with a comment.

        """
        router = MetadataRouter(owner=self.__class__.__name__)

        # first we add all steps except the last one
        for _, name, trans in self._iter(with_final=False, filter_passthrough=True):
            method_mapping = MethodMapping()
            # fit, fit_predict, and fit_transform call fit_transform if it
            # exists, or else fit and transform
            if hasattr(trans, "fit_transform"):
                (
                    method_mapping.add(caller="fit", callee="fit_transform")
                    .add(caller="fit_transform", callee="fit_transform")
                    .add(caller="fit_predict", callee="fit_transform")
                )
            else:
                (
                    method_mapping.add(caller="fit", callee="fit")
                    .add(caller="fit", callee="transform")
                    .add(caller="fit_transform", callee="fit")
                    .add(caller="fit_transform", callee="transform")
                    .add(caller="fit_predict", callee="fit")
                    .add(caller="fit_predict", callee="transform")
                )

            (
                method_mapping.add(caller="predict", callee="transform")
                .add(caller="predict", callee="transform")
                .add(caller="predict_proba", callee="transform")
                .add(caller="decision_function", callee="transform")
                .add(caller="predict_log_proba", callee="transform")
                .add(caller="transform", callee="transform")
                .add(caller="inverse_transform", callee="inverse_transform")
                .add(caller="score", callee="transform")
            )

            router.add(method_mapping=method_mapping, **{name: trans})

        # Only the _non_post_processing_steps is changed from the original
        # implementation is changed in the following line
        final_name, final_est = self._non_post_processing_steps()[-1]
        if final_est is None or final_est == "passthrough":
            return router

        # then we add the last step
        method_mapping = MethodMapping()
        if hasattr(final_est, "fit_transform"):
            method_mapping.add(caller="fit_transform", callee="fit_transform")
        else:
            method_mapping.add(caller="fit", callee="fit").add(
                caller="fit",
                callee="transform",
            )
        (
            method_mapping.add(caller="fit", callee="fit")
            .add(caller="predict", callee="predict")
            .add(caller="fit_predict", callee="fit_predict")
            .add(caller="predict_proba", callee="predict_proba")
            .add(caller="decision_function", callee="decision_function")
            .add(caller="predict_log_proba", callee="predict_log_proba")
            .add(caller="transform", callee="transform")
            .add(caller="inverse_transform", callee="inverse_transform")
            .add(caller="score", callee="score")
        )

        router.add(method_mapping=method_mapping, **{final_name: final_est})
        return router



"""Contains functions for molecular similarity."""

import numpy as np
import numpy.typing as npt
from scipy import sparse


def tanimoto_similarity_sparse(
    matrix_a: sparse.csr_matrix | npt.NDArray[np.int_],
    matrix_b: sparse.csr_matrix | npt.NDArray[np.int_],
) -> npt.NDArray[np.float64]:
    """Calculate a matrix of tanimoto similarities between feature matrix a and b.

    Parameters
    ----------
    matrix_a : sparse.csr_matrix
        Feature matrix A.
    matrix_b : sparse.csr_matrix
        Feature matrix B.

    Returns
    -------
    npt.NDArray[np.float64]
        Matrix of similarity values between instances of A (rows/first dim) ,
        and instances of B (columns/second dim).

    """
    if isinstance(matrix_a, np.ndarray):
        matrix_a = sparse.csr_matrix(matrix_a)
    if isinstance(matrix_b, np.ndarray):
        matrix_b = sparse.csr_matrix(matrix_b)
    intersection = matrix_a.dot(matrix_b.transpose()).toarray()
    norm_1 = np.array(matrix_a.sum(axis=1))
    norm_2 = norm_1 if matrix_a is matrix_b else np.array(matrix_b.sum(axis=1))
    union = norm_1 + norm_2.T - intersection
    # avoid division by zero https://stackoverflow.com/a/37977222
    return np.divide(
        intersection,
        union,
        out=np.zeros(intersection.shape, dtype=float),
        where=union != 0,
    )


def tanimoto_distance_sparse(
    matrix_a: sparse.csr_matrix | npt.NDArray[np.int_],
    matrix_b: sparse.csr_matrix | npt.NDArray[np.int_],
) -> npt.NDArray[np.float64]:
    """Calculate a matrix of tanimoto distance between feature matrix a and b.

    Tanimoto distance is defined as 1-similarity.

    Parameters
    ----------
    matrix_a : sparse.csr_matrix | npt.NDArray[np.int_]
        Feature matrix A.
    matrix_b : sparse.csr_matrix | npt.NDArray[np.int_]
        Feature matrix B.

    Returns
    -------
    npt.NDArray[np.float64]
        Matrix of similarity values between instances of A (rows/first dim)
        and instances of B (columns/second dim).

    """
    return 1 - tanimoto_similarity_sparse(matrix_a, matrix_b)  # type: ignore


def pairwise_tanimoto_similarity(
    matrix_a: sparse.csr_matrix | npt.NDArray[np.int_],
    matrix_b: sparse.csr_matrix | npt.NDArray[np.int_],
) -> np.float64:
    """Calculate the tanimoto similarity between singles rows of feature matrix a and b.

    This function is required for sklean metrics which expect a single similarity value
    between two instances. The input matrices must have exactly one row.

    Parameters
    ----------
    matrix_a : sparse.csr_matrix | npt.NDArray[np.int_]
        Feature matrix A.
    matrix_b : sparse.csr_matrix | npt.NDArray[np.int_]
        Feature matrix B.

    Returns
    -------
    np.float64
        Tanimoto similarity between the two rows.

    Raises
    ------
    ValueError
        If either matrix does not have exactly one row.

    """
    if isinstance(matrix_a, np.ndarray):
        matrix_a = sparse.csr_matrix(matrix_a)
    if isinstance(matrix_b, np.ndarray):
        matrix_b = sparse.csr_matrix(matrix_b)
    if matrix_a.shape[0] != 1 or matrix_b.shape[0] != 1:
        raise ValueError("Both matrices must have exactly one row.")
    return tanimoto_similarity_sparse(matrix_a, matrix_b)[0, 0]


def pairwise_tanimoto_distance(
    matrix_a: sparse.csr_matrix | npt.NDArray[np.int_],
    matrix_b: sparse.csr_matrix | npt.NDArray[np.int_],
) -> np.float64:
    """Calculate the tanimoto distance between singles rows of feature matrix a and b.

    Tanimoto distance is defined as 1-similarity.


    This function is required for sklean metrics which expect a single similarity value
    between two instances. The input matrices must have exactly one row.

    Parameters
    ----------
    matrix_a : sparse.csr_matrix | npt.NDArray[np.int_]
        Feature matrix A.
    matrix_b : sparse.csr_matrix | npt.NDArray[np.int_]
        Feature matrix B.

    Returns
    -------
    np.float64
        Tanimoto distance between the two rows.

    """
    return 1 - pairwise_tanimoto_similarity(matrix_a, matrix_b)


def self_tanimoto_similarity(
    matrix_a: sparse.csr_matrix | npt.NDArray[np.int_],
) -> npt.NDArray[np.float64]:
    """Calculate a matrix of tanimoto similarity between feature matrix a and itself.

    Parameters
    ----------
    matrix_a : Union[sparse.csr_matrix, npt.NDArray[np.int_]]
        Feature matrix.

    Returns
    -------
    npt.NDArray[np.float64]
        Square matrix of similarity values between all instances in the matrix.

    Raises
    ------
    TypeError
        If the matrix is not a sparse matrix or a numpy array.

    """
    if isinstance(matrix_a, np.ndarray):
        sparse_matrix = sparse.csr_matrix(matrix_a)
    elif isinstance(matrix_a, sparse.csr_matrix):
        sparse_matrix = matrix_a
    else:
        raise TypeError(f"Unsupported type: {type(matrix_a)}")
    return tanimoto_similarity_sparse(sparse_matrix, sparse_matrix)


def self_tanimoto_distance(
    matrix_a: sparse.csr_matrix | npt.NDArray[np.int_],
) -> npt.NDArray[np.float64]:
    """Calculate a matrix of tanimoto distance between feature matrix a and itself.

    Parameters
    ----------
    matrix_a : Union[sparse.csr_matrix, npt.NDArray[np.int_]]
        Feature matrix.

    Returns
    -------
    npt.NDArray[np.float64]
        Square matrix of similarity values between all instances in the matrix.

    """
    return 1 - self_tanimoto_similarity(matrix_a)  # type: ignore



"""Tests for the MolToPathFingerprint pipeline element."""

import unittest
from typing import Any

import numpy as np
from sklearn import clone

from molpipeline import Pipeline
from molpipeline.any2mol import SmilesToMol
from molpipeline.mol2any import Mol2PathFP

# pylint: disable=duplicate-code

test_smiles = [
    "c1ccccc1",
    "c1ccccc1C",
    "NCCOCCCC(=O)O",
]


class TestMol2PathFingerprint(unittest.TestCase):
    """Unittest for Mol2PathFP, which calculates the RDKit Path Fingerprint."""

    def test_clone(self) -> None:
        """Test if the Mol2PathFP pipeline element can be constructed."""
        mol_fp = Mol2PathFP()
        mol_fp_copy = clone(mol_fp)
        self.assertTrue(mol_fp_copy is not mol_fp)
        for key, value in mol_fp.get_params().items():
            self.assertEqual(value, mol_fp_copy.get_params()[key])
        mol_fp_recreated = Mol2PathFP(**mol_fp.get_params())
        for key, value in mol_fp.get_params().items():
            self.assertEqual(value, mol_fp_recreated.get_params()[key])

    def test_output_types(self) -> None:
        """Test equality of different output_types."""
        smi2mol = SmilesToMol()
        sparse_path_fp = Mol2PathFP(n_bits=1024, return_as="sparse")
        dense_path_fp = Mol2PathFP(n_bits=1024, return_as="dense")
        rdkit_vect_path_fp = Mol2PathFP(n_bits=1024, return_as="rdkit")
        sparse_pipeline = Pipeline(
            [
                ("smi2mol", smi2mol),
                ("sparse_path_fp", sparse_path_fp),
            ],
        )
        dense_pipeline = Pipeline(
            [
                ("smi2mol", smi2mol),
                ("dense_path_fp", dense_path_fp),
            ],
        )
        rdkit_vect_pipeline = Pipeline(
            [
                ("smi2mol", smi2mol),
                ("rdkit_vect_path_fp", rdkit_vect_path_fp),
            ],
        )

        sparse_output = sparse_pipeline.fit_transform(test_smiles)
        dense_output = dense_pipeline.fit_transform(test_smiles)
        rdkit_vect_path_fp_output = rdkit_vect_pipeline.fit_transform(test_smiles)

        self.assertTrue(np.all(sparse_output.toarray() == dense_output))

        self.assertTrue(
            np.equal(
                dense_output,
                np.array(rdkit_vect_path_fp_output),
            ).all(),
        )

    def test_counted_bits(self) -> None:
        """Test if the option counted bits works as expected."""
        mol_fp = Mol2PathFP(n_bits=1024, return_as="dense")
        smi2mol = SmilesToMol()
        pipeline = Pipeline(
            [
                ("smi2mol", smi2mol),
                ("mol_fp", mol_fp),
            ],
        )
        output_binary = pipeline.fit_transform(test_smiles)
        pipeline.set_params(mol_fp__counted=True)
        output_counted = pipeline.fit_transform(test_smiles)
        self.assertTrue(
            np.all(np.flatnonzero(output_counted) == np.flatnonzero(output_binary)),
        )
        self.assertTrue(np.all(output_counted >= output_binary))
        self.assertTrue(np.any(output_counted > output_binary))

    def test_setter_getter(self) -> None:
        """Test if the setters and getters work as expected."""
        mol_fp = Mol2PathFP()
        params: dict[str, Any] = {
            "min_path": 10,
            "max_path": 12,
            "use_hs": False,
            "branched_paths": False,
            "use_bond_order": False,
            "count_simulation": True,
            "num_bits_per_feature": 4,
            "counted": True,
            "n_bits": 1024,
        }
        mol_fp.set_params(**params)
        self.assertEqual(mol_fp.get_params()["min_path"], 10)
        self.assertEqual(mol_fp.get_params()["max_path"], 12)
        self.assertEqual(mol_fp.get_params()["use_hs"], False)
        self.assertEqual(mol_fp.get_params()["branched_paths"], False)
        self.assertEqual(mol_fp.get_params()["use_bond_order"], False)
        self.assertEqual(mol_fp.get_params()["count_simulation"], True)
        self.assertEqual(mol_fp.get_params()["num_bits_per_feature"], 4)
        self.assertEqual(mol_fp.get_params()["counted"], True)
        self.assertEqual(mol_fp.get_params()["n_bits"], 1024)

    def test_setter_getter_error_handling(self) -> None:
        """Test the setters and getters work as expected when errors are encountered."""
        mol_fp = Mol2PathFP()
        params: dict[str, Any] = {
            "min_path": 2,
            "n_bits": 1024,
            "return_as": "invalid-option",
        }
        self.assertRaises(ValueError, mol_fp.set_params, **params)

    def test_feature_names(self) -> None:
        """Test if the feature names are correct."""
        mol_fp = Mol2PathFP(n_bits=1024)
        feature_names = mol_fp.feature_names
        self.assertEqual(len(feature_names), 1024)
        # feature names should be unique
        self.assertEqual(len(feature_names), len(set(feature_names)))


if __name__ == "__main__":
    unittest.main()



"""Methods for calibrating predicted probabilities."""

# Authors: The scikit-learn developers
# SPDX-License-Identifier: BSD-3-Clause
# Adapted by: MolPipeline developers

import warnings
from inspect import signature
from typing import Any, Literal, Self

import numpy as np
import numpy.typing as npt
from loguru import logger
from sklearn.base import (
    BaseEstimator,
    _fit_context,  # noqa: PLC2701
    clone,
)
from sklearn.calibration import (
    CalibratedClassifierCV as SklearnCalibratedClassifierCV,
)
from sklearn.calibration import (
    _CalibratedClassifier,
    _fit_calibrator,  # noqa: PLC2701
)
from sklearn.frozen import FrozenEstimator
from sklearn.model_selection import LeaveOneOut, check_cv, cross_val_predict
from sklearn.preprocessing import LabelEncoder
from sklearn.utils import (
    Bunch,
    _safe_indexing,  # noqa: PLC2701
    indexable,
)
from sklearn.utils._array_api import (  # noqa: PLC2701
    get_namespace_and_device,
    move_to,
)
from sklearn.utils._response import (  # noqa: PLC2701
    _get_response_values,
    _process_predict_proba,
)
from sklearn.utils.class_weight import compute_sample_weight
from sklearn.utils.metadata_routing import (
    _routing_enabled,  # noqa: PLC2701
    process_routing,
)
from sklearn.utils.multiclass import check_classification_targets
from sklearn.utils.parallel import Parallel, delayed
from sklearn.utils.validation import (
    _check_method_params,  # noqa: PLC2701
    _check_response_method,  # noqa: PLC2701
    _check_sample_weight,  # noqa: PLC2701
)
from typing_extensions import override


def merge_class_and_sample_weights(
    y: npt.ArrayLike,
    class_weight: dict[Any, float] | Literal["balanced"] | None,
    sample_weight: npt.ArrayLike | None,
) -> npt.ArrayLike | None:
    """Merge class weights and sample weights into a single sample weight array.

    Parameters
    ----------
    y : array-like of shape (n_samples,)
        Target values.

    class_weight : dict[Any, float] | 'balanced' | None, optional
        Class weights used for the calibration.
        If a dict, it must be provided in this form: ``{class_label: weight}``.
        Those weights won't be used for the underlying estimator training.
        See :term:`Glossary <class_weight>` for more details.

    sample_weight : npt.ArrayLike | None, optional
        Sample weights. If None, then samples are equally weighted.

    Returns
    -------
    merged_sample_weight : ndarray of shape (n_samples,) or None
        Merged sample weights. If both `class_weight` and `sample_weight` are None,
        returns None.

    """
    if class_weight is None:
        return sample_weight

    class_weights_array = compute_sample_weight(class_weight, y)
    if sample_weight is None:
        return class_weights_array

    return class_weights_array * sample_weight


def _fit_classifier_calibrator_pair(  # noqa: PLR0917  # pylint: disable=R0914,R0917
    estimator: BaseEstimator,
    X: npt.ArrayLike,  # noqa: N803
    y: npt.ArrayLike,
    train: npt.NDArray[np.int_],
    test: npt.NDArray[np.int_],
    method: Literal["sigmoid", "isotonic", "temperature"],
    classes: npt.ArrayLike,
    xp: Any,
    class_weight: dict[Any, float] | Literal["balanced"] | None = None,
    sample_weight: npt.ArrayLike | None = None,
    fit_params: dict[str, Any] | None = None,
) -> _CalibratedClassifier:
    """Fit a classifier/calibration pair on a given train/test split.

    Adapted from scikit-learn to add support for class weights during calibration.

    Fit the classifier on the train set, compute its predictions on the test
    set and use the predictions as input to fit the calibrator along with the
    test labels.

    Parameters
    ----------
    estimator : estimator instance
        Cloned base estimator.

    X : array-like, shape (n_samples, n_features)
        Sample data.

    y : array-like, shape (n_samples,)
        Targets.

    train : ndarray, shape (n_train_indices,)
        Indices of the training subset.

    test : ndarray, shape (n_test_indices,)
        Indices of the testing subset.

    method : {'sigmoid', 'isotonic', 'temperature'}
        Method to use for calibration.

    classes : ndarray, shape (n_classes,)
        The target classes.

    xp : namespace
        Array API namespace.

    class_weight : dict or 'balanced', optional
        Class weights used for the calibration.
        If a dict, it must be provided in this form: ``{class_label: weight}``.
        Those weights won't be used for the underlying estimator training.
        See :term:`Glossary <class_weight>` for more details.

    sample_weight : array-like, default=None
        Sample weights for `X`.

    fit_params : dict, default=None
        Parameters to pass to the `fit` method of the underlying
        classifier.

    Returns
    -------
    calibrated_classifier : _CalibratedClassifier instance

    """
    fit_params_train = _check_method_params(X, params=fit_params, indices=train)
    X_train, y_train = _safe_indexing(X, train), _safe_indexing(y, train)  # noqa: N806
    X_test, y_test = _safe_indexing(X, test), _safe_indexing(y, test)  # noqa: N806

    estimator.fit(X_train, y_train, **fit_params_train)

    predictions, _ = _get_response_values(  # pylint: disable=W0632
        estimator,
        X_test,
        response_method=["decision_function", "predict_proba"],
    )
    if predictions.ndim == 1:
        # Reshape binary output from `(n_samples,)` to `(n_samples, 1)`
        predictions = predictions.reshape(-1, 1)

    if sample_weight is not None:
        # Check that the sample_weight dtype is consistent with the predictions
        # to avoid unintentional upcasts.
        sample_weight = _check_sample_weight(sample_weight, X, dtype=predictions.dtype)
        sw_test = _safe_indexing(sample_weight, test)
    else:
        sw_test = None
    sw_test = merge_class_and_sample_weights(y_test, class_weight, sw_test)
    return _fit_calibrator(
        estimator,
        predictions,
        y_test,
        classes,
        method,
        xp=xp,
        sample_weight=sw_test,
    )


class CalibratedClassifierCV(SklearnCalibratedClassifierCV):
    """Calibrate probabilities using isotonic, sigmoid, or temperature scaling.

    This class is adapted from scikit-learn to add support for class weights during
    calibration. In addtion code was adapted to adhere to local style guides.
    The plan is to upstream these changes to scikit-learn in the future, once a
    rigerous study has been performed to validate the approach.

    This class uses cross-validation to both estimate the parameters of a
    classifier and subsequently calibrate a classifier. With
    `ensemble=True`, for each cv split it
    fits a copy of the base estimator to the training subset, and calibrates it
    using the testing subset. For prediction, predicted probabilities are
    averaged across these individual calibrated classifiers. When
    `ensemble=False`, cross-validation is used to obtain unbiased predictions,
    via :func:`~sklearn.model_selection.cross_val_predict`, which are then
    used for calibration. For prediction, the base estimator, trained using all
    the data, is used. This is the prediction method implemented when
    `probabilities=True` for :class:`~sklearn.svm.SVC` and :class:`~sklearn.svm.NuSVC`
    estimators (see :ref:`User Guide <scores_probabilities>` for details).

    Already fitted classifiers can be calibrated by wrapping the model in a
    :class:`~sklearn.frozen.FrozenEstimator`. In this case all provided
    data is used for calibration. The user has to take care manually that data
    for model fitting and calibration are disjoint.

    The calibration is based on the :term:`decision_function` method of the
    `estimator` if it exists, else on :term:`predict_proba`.

    Read more in the :ref:`User Guide <calibration>`.
    In order to learn more on the CalibratedClassifierCV class, see the
    following calibration examples:
    :ref:`sphx_glr_auto_examples_calibration_plot_calibration.py`,
    :ref:`sphx_glr_auto_examples_calibration_plot_calibration_curve.py`, and
    :ref:`sphx_glr_auto_examples_calibration_plot_calibration_multiclass.py`.

    Parameters
    ----------
    estimator : estimator instance, default=None
        The classifier whose output need to be calibrated to provide more
        accurate `predict_proba` outputs. The default classifier is
        a :class:`~sklearn.svm.LinearSVC`.

        .. versionadded:: 1.2

    method : {'sigmoid', 'isotonic', 'temperature'}, default='sigmoid'
        The method to use for calibration. Can be:

        - 'sigmoid', which corresponds to Platt's method (i.e. a binary logistic
          regression model).
        - 'isotonic', which is a non-parametric approach.
        - 'temperature', temperature scaling.

        Sigmoid and isotonic calibration methods natively support only binary
        classifiers and extend to multi-class classification using a One-vs-Rest (OvR)
        strategy with post-hoc renormalization, i.e., adjusting the probabilities after
        calibration to ensure they sum up to 1.

        In contrast, temperature scaling naturally supports multi-class calibration by
        applying `softmax(classifier_logits/T)` with a value of `T` (temperature)
        that optimizes the log loss.

        For very uncalibrated classifiers on very imbalanced datasets, sigmoid
        calibration might be preferred because it fits an additional intercept
        parameter. This helps shift decision boundaries appropriately when the
        classifier being calibrated is biased towards the majority class.

        Isotonic calibration is not recommended when the number of calibration samples
        is too low ``(≪1000)`` since it then tends to overfit.

        .. versionchanged:: 1.8
           Added option 'temperature'.

    cv : int, cross-validation generator, or iterable, default=None
        Determines the cross-validation splitting strategy.
        Possible inputs for cv are:

        - None, to use the default 5-fold cross-validation,
        - integer, to specify the number of folds.
        - :term:`CV splitter`,
        - An iterable yielding (train, test) splits as arrays of indices.

        For integer/None inputs, if ``y`` is binary or multiclass,
        :class:`~sklearn.model_selection.StratifiedKFold` is used. If ``y`` is
        neither binary nor multiclass, :class:`~sklearn.model_selection.KFold`
        is used.

        Refer to the :ref:`User Guide <cross_validation>` for the various
        cross-validation strategies that can be used here.

        .. versionchanged:: 0.22
            ``cv`` default value if None changed from 3-fold to 5-fold.

    n_jobs : int, default=None
        Number of jobs to run in parallel.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors.

        Base estimator clones are fitted in parallel across cross-validation
        iterations.

        See :term:`Glossary <n_jobs>` for more details.

        .. versionadded:: 0.24

    ensemble : bool, or "auto", default="auto"
        Determines how the calibrator is fitted.

        "auto" will use `False` if the `estimator` is a
        :class:`~sklearn.frozen.FrozenEstimator`, and `True` otherwise.

        If `True`, the `estimator` is fitted using training data, and
        calibrated using testing data, for each `cv` fold. The final estimator
        is an ensemble of `n_cv` fitted classifier and calibrator pairs, where
        `n_cv` is the number of cross-validation folds. The output is the
        average predicted probabilities of all pairs.

        If `False`, `cv` is used to compute unbiased predictions, via
        :func:`~sklearn.model_selection.cross_val_predict`, which are then
        used for calibration. At prediction time, the classifier used is the
        `estimator` trained on all the data.
        Note that this method is also internally implemented  in
        :mod:`sklearn.svm` estimators with the `probabilities=True` parameter.

        .. versionadded:: 0.24

        .. versionchanged:: 1.6
            `"auto"` option is added and is the default.

    class_weight : dict or 'balanced', default=None
        Class weights used for the calibration.
        If a dict, it must be provided in this form: ``{class_label: weight}``.
        Those weights won't be used for the underlying estimator training.
        See :term:`Glossary <class_weight>` for more details.
        If both `class_weight` and `sample_weight` are provided during
        :meth:`fit`, the sample weights are multiplied by the class weights.

    Attributes
    ----------
    classes_ : ndarray of shape (n_classes,)
        The class labels.

    n_features_in_ : int
        Number of features seen during :term:`fit`. Only defined if the
        underlying estimator exposes such an attribute when fit.

        .. versionadded:: 0.24

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of features seen during :term:`fit`. Only defined if the
        underlying estimator exposes such an attribute when fit.

        .. versionadded:: 1.0

    calibrated_classifiers_ : list (len() equal to cv or 1 if `ensemble=False`)
        The list of classifier and calibrator pairs.

        - When `ensemble=True`, `n_cv` fitted `estimator` and calibrator pairs.
          `n_cv` is the number of cross-validation folds.
        - When `ensemble=False`, the `estimator`, fitted on all the data, and fitted
          calibrator.

        .. versionchanged:: 0.24
            Single calibrated classifier case when `ensemble=False`.

    See Also
    --------
    calibration_curve : Compute true and predicted probabilities
        for a calibration curve.

    References
    ----------
    .. [1] B. Zadrozny & C. Elkan.
       `Obtaining calibrated probability estimates from decision trees
       and naive Bayesian classifiers
       <https://cseweb.ucsd.edu/~elkan/calibrated.pdf>`_, ICML 2001.

    .. [2] B. Zadrozny & C. Elkan.
       `Transforming Classifier Scores into Accurate Multiclass
       Probability Estimates
       <https://web.archive.org/web/20060720141520id_/http://www.research.ibm.com:80/people/z/zadrozny/kdd2002-Transf.pdf>`_,
       KDD 2002.

    .. [3] J. Platt. `Probabilistic Outputs for Support Vector Machines
       and Comparisons to Regularized Likelihood Methods
       <https://www.researchgate.net/profile/John-Platt-2/publication/2594015_Probabilistic_Outputs_for_Support_Vector_Machines_and_Comparisons_to_Regularized_Likelihood_Methods/links/004635154cff5262d6000000/Probabilistic-Outputs-for-Support-Vector-Machines-and-Comparisons-to-Regularized-Likelihood-Methods.pdf>`_,
       1999.

    .. [4] A. Niculescu-Mizil & R. Caruana.
       `Predicting Good Probabilities with Supervised Learning
       <https://www.cs.cornell.edu/~alexn/papers/calibration.icml05.crc.rev3.pdf>`_,
       ICML 2005.

    .. [5] Chuan Guo, Geoff Pleiss, Yu Sun, Kilian Q. Weinberger.
       :doi:`On Calibration of Modern Neural Networks<10.48550/arXiv.1706.04599>`.
       Proceedings of the 34th International Conference on Machine Learning,
       PMLR 70:1321-1330, 2017.

    Examples
    --------
    >>> from sklearn.datasets import make_classification
    >>> from sklearn.naive_bayes import GaussianNB
    >>> from sklearn.calibration import CalibratedClassifierCV
    >>> X, y = make_classification(n_samples=100, n_features=2,
    ...                            n_redundant=0, random_state=42)
    >>> base_clf = GaussianNB()
    >>> calibrated_clf = CalibratedClassifierCV(base_clf, cv=3)
    >>> calibrated_clf.fit(X, y)
    CalibratedClassifierCV(...)
    >>> len(calibrated_clf.calibrated_classifiers_)
    3
    >>> calibrated_clf.predict_proba(X)[:5, :]
    array([[0.110, 0.889],
           [0.072, 0.927],
           [0.928, 0.072],
           [0.928, 0.072],
           [0.072, 0.928]])
    >>> from sklearn.model_selection import train_test_split
    >>> X, y = make_classification(n_samples=100, n_features=2,
    ...                            n_redundant=0, random_state=42)
    >>> X_train, X_calib, y_train, y_calib = train_test_split(
    ...        X, y, random_state=42
    ... )
    >>> base_clf = GaussianNB()
    >>> base_clf.fit(X_train, y_train)
    GaussianNB()
    >>> from sklearn.frozen import FrozenEstimator
    >>> calibrated_clf = CalibratedClassifierCV(FrozenEstimator(base_clf))
    >>> calibrated_clf.fit(X_calib, y_calib)
    CalibratedClassifierCV(...)
    >>> len(calibrated_clf.calibrated_classifiers_)
    1
    >>> calibrated_clf.predict_proba([[-0.5, 0.5]])
    array([[0.936, 0.063]])

    """

    class_weight: dict[Any, float] | Literal["balanced"] | None

    def __init__(
        self,
        estimator: BaseEstimator | None = None,
        *,
        method: Literal["sigmoid", "isotonic", "temperature"] = "sigmoid",
        cv: Any = None,
        n_jobs: int | None = None,
        ensemble: Literal["auto"] | bool = "auto",
        class_weight: dict[Any, float] | Literal["balanced"] | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the CalibratedClassifierCV instance.

        Parameters
        ----------
        estimator : BaseEstimator | None, optional
            The classifier whose output need to be calibrated to provide more
            accurate `predict_proba` outputs. The default classifier is
            a :class:`~sklearn.svm.LinearSVC`.
        method : Literal['sigmoid', 'isotonic', 'temperature'], default='sigmoid'
            The method to use for calibration. Can be:
            - 'sigmoid', which corresponds to Platt's method (i.e. A binary logistic
              regression model)
            - 'isotonic', which is a non-parametric approach
            - 'temperature', temperature scaling
        cv : int, cross-validation generator, or iterable, optional
            Determines the cross-validation splitting strategy.
        n_jobs : int, optional
            Number of jobs to run in parallel.
            ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
            The value ``-1`` means using all processors.
        ensemble : Literal["auto"] | bool, default="auto"
            Determines how the calibrator is fitted.
            The value "auto" will use `False` if the `estimator` is a
            :class:`~sklearn.frozen.FrozenEstimator`, and `True` otherwise.
        class_weight : dict | Literal['balanced'], optional
            Class weights used for the calibration.
            If a dict, it must be provided in this form: ``{class_label: weight}``.
            Those weights won't be used for the underlying estimator training.
            See :term:`Glossary <class_weight>` for more details.
        **kwargs : Any
            Addition parameters set via ``set_params``.

        """
        if method == "temperature" and class_weight is not None:
            logger.warning(
                "At the moment temperature scaling with class weights "
                "seems to have no effect.",
            )
        super().__init__(
            estimator=estimator,
            method=method,
            cv=cv,
            n_jobs=n_jobs,
            ensemble=ensemble,
        )
        self.class_weight = class_weight
        self.set_params(**kwargs)

    @override
    @_fit_context(
        # CalibratedClassifierCV.estimator is not validated yet
        prefer_skip_nested_validation=False,
    )
    def fit(  # noqa: PLR0912, PLR0914, PLR0915  # pylint: disable=R0912,R0914,R0915
        self,
        X: npt.ArrayLike,
        y: npt.ArrayLike,
        sample_weight: npt.ArrayLike | None = None,
        **fit_params: Any,
    ) -> Self:
        """Fit the calibrated model.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data.

        y : array-like of shape (n_samples,)
            Target values.

        sample_weight : array-like of shape (n_samples,), default=None
            Sample weights. If None, then samples are equally weighted.

        **fit_params : dict
            Parameters to pass to the `fit` method of the underlying
            classifier.

        Returns
        -------
        self : object
            Returns an instance of self.

        Raises
        ------
        ValueError
            If `cv` is LeaveOneOut, because a model cannot be calibrated on a single
            sample.

        """
        check_classification_targets(y)
        X, y = indexable(X, y)  # noqa: N806
        estimator = self._get_estimator()

        ensemble = self.ensemble
        if ensemble == "auto":
            ensemble = not isinstance(estimator, FrozenEstimator)

        self.calibrated_classifiers_ = []  # pylint: disable=W0201

        # Set `classes_` using all `y`
        label_encoder_ = LabelEncoder().fit(y)
        self.classes_ = label_encoder_.classes_  # pylint: disable=W0201
        if self.method == "temperature" and isinstance(y[0], str):  # type: ignore
            # for temperature scaling if `y` contains strings then encode it
            # right here to avoid fitting LabelEncoder again within the
            # `_fit_calibrator` function.
            y = label_encoder_.transform(y=y)

        if _routing_enabled():
            routed_params = process_routing(
                self,
                "fit",
                sample_weight=sample_weight,
                **fit_params,
            )
        else:
            # sample_weight checks
            fit_parameters = signature(estimator.fit).parameters
            supports_sw = "sample_weight" in fit_parameters
            if sample_weight is not None and not supports_sw:
                estimator_name = type(estimator).__name__
                warnings.warn(
                    f"Since {estimator_name} does not appear to accept"
                    " sample_weight, sample weights will only be used for the"
                    " calibration itself. This can be caused by a limitation of"
                    " the current scikit-learn API. See the following issue for"
                    " more details:"
                    " https://github.com/scikit-learn/scikit-learn/issues/21134."
                    " Be warned that the result of the calibration is likely to be"
                    " incorrect.",
                    stacklevel=2,
                )
            routed_params = Bunch()
            routed_params.splitter = Bunch(split={})  # no routing for splitter
            routed_params.estimator = Bunch(fit=fit_params)
            if sample_weight is not None and supports_sw:
                routed_params.estimator.fit["sample_weight"] = sample_weight

        xp, is_array_api, device_ = get_namespace_and_device(X)
        if is_array_api:
            y, sample_weight = move_to(  # type: ignore  # pylint: disable=W0632
                y,
                sample_weight,
                xp=xp,
                device=device_,
            )
        # Check that each cross-validation fold can have at least one
        # example per class
        if isinstance(self.cv, int):
            n_folds = self.cv
        elif hasattr(self.cv, "n_splits"):
            n_folds = self.cv.n_splits
        else:
            n_folds = None
        if n_folds and xp.any(xp.unique_counts(y)[1] < n_folds):
            raise ValueError(
                f"Requesting {n_folds}-fold "
                "cross-validation but provided less than "
                f"{n_folds} examples for at least one class.",
            )
        if isinstance(self.cv, LeaveOneOut):
            raise ValueError(
                "LeaveOneOut cross-validation does not allow"
                "all classes to be present in test splits. "
                "Please use a cross-validation generator that allows "
                "all classes to appear in every test and train split.",
            )
        cv = check_cv(self.cv, y, classifier=True)

        if ensemble:
            parallel = Parallel(n_jobs=self.n_jobs)
            self.calibrated_classifiers_ = parallel(  # pylint: disable=W0201
                delayed(_fit_classifier_calibrator_pair)(
                    clone(estimator),
                    X,
                    y,
                    train=train,
                    test=test,
                    method=self.method,
                    classes=self.classes_,
                    class_weight=self.class_weight,
                    sample_weight=sample_weight,
                    fit_params=routed_params.estimator.fit,
                    xp=xp,
                )
                for train, test in cv.split(X, y, **routed_params.splitter.split)
            )
        else:
            calibration_sample_weight = merge_class_and_sample_weights(
                y,
                self.class_weight,
                sample_weight,
            )
            this_estimator = clone(estimator)
            method_name = _check_response_method(
                this_estimator,
                ["decision_function", "predict_proba"],
            ).__name__
            predictions = cross_val_predict(
                estimator=this_estimator,
                X=X,
                y=y,
                cv=cv,
                method=method_name,
                n_jobs=self.n_jobs,
                params=routed_params.estimator.fit,
            )
            if self.classes_.shape[0] == 2:  # noqa: PLR2004
                # Ensure shape (n_samples, 1) in the binary case
                if method_name == "predict_proba":
                    # Select the probability column of the positive class
                    predictions = _process_predict_proba(
                        y_pred=predictions,
                        target_type="binary",
                        classes=self.classes_,
                        pos_label=self.classes_[1],
                    )
                predictions = predictions.reshape(-1, 1)

            if calibration_sample_weight is not None:
                # Check that the sample_weight dtype is consistent with the
                # predictions to avoid unintentional upcasts.
                calibration_sample_weight = _check_sample_weight(
                    calibration_sample_weight,
                    predictions,
                    dtype=predictions.dtype,
                )

            this_estimator.fit(X, y, **routed_params.estimator.fit)
            # Note: Here we don't pass on fit_params because the supported
            # calibrators don't support fit_params anyway
            calibrated_classifier = _fit_calibrator(
                this_estimator,
                predictions,
                y,
                self.classes_,
                self.method,
                sample_weight=calibration_sample_weight,
                xp=xp,
            )
            self.calibrated_classifiers_.append(calibrated_classifier)

        first_clf = self.calibrated_classifiers_[0].estimator
        if hasattr(first_clf, "n_features_in_"):
            self.n_features_in_ = first_clf.n_features_in_  # pylint: disable=W0201
        if hasattr(first_clf, "feature_names_in_"):
            self.feature_names_in_ = first_clf.feature_names_in_  # pylint: disable=W0201
        return self



"""Tests for the MolToElementCount pipeline element."""

import unittest

import numpy as np
from rdkit.Chem import MolFromSmiles
from sklearn import clone

from molpipeline import Pipeline
from molpipeline.any2mol import SmilesToMol
from molpipeline.mol2any import MolToElementCount


class TestMolToElementCount(unittest.TestCase):
    """Unittest for MolToElementCount."""

    def test_default_construction(self) -> None:
        """Test default construction with all 118 elements."""
        element_count = MolToElementCount(standardizer=None)
        self.assertEqual(element_count.n_features, 118)
        self.assertEqual(len(element_count.feature_names), 118)
        self.assertEqual(element_count.feature_names[0], "Count_H")
        self.assertEqual(element_count.feature_names[5], "Count_C")

    def test_custom_element_list(self) -> None:
        """Test construction with a custom element list."""
        element_count = MolToElementCount(element_list=[1, 6, 7, 8], standardizer=None)
        self.assertEqual(element_count.n_features, 4)
        self.assertEqual(
            element_count.feature_names,
            ["Count_H", "Count_C", "Count_N", "Count_O"],
        )

    def test_invalid_element_list(self) -> None:
        """Test that invalid element lists raise ValueError."""
        with self.assertRaises(ValueError):
            MolToElementCount(element_list=[0, 1])
        with self.assertRaises(ValueError):
            MolToElementCount(element_list=[119])
        with self.assertRaises(ValueError):
            MolToElementCount(element_list=[-1])

    def test_pretransform_single(self) -> None:
        """Test element counting for a single molecule.

        Raises
        ------
        AssertionError
            If pretransform_single does not return an np.ndarray.

        """
        element_count = MolToElementCount(element_list=[6, 8], standardizer=None)
        mol = MolFromSmiles("CCO")  # ethanol: 2 C, 1 O
        result = element_count.pretransform_single(mol)
        if not isinstance(result, np.ndarray):
            raise AssertionError(f"Expected np.ndarray, got {type(result)}")
        self.assertTrue(np.array_equal(result, [2.0, 1.0]))
        self.assertEqual(result.dtype, np.float64)

        element_count = MolToElementCount(element_list=[8], standardizer=None)
        mol = MolFromSmiles("CCO")  # ethanol: 2 C, 1 O
        result = element_count.pretransform_single(mol)
        if not isinstance(result, np.ndarray):
            raise AssertionError(f"Expected np.ndarray, got {type(result)}")
        self.assertTrue(np.array_equal(result, [1.0]))
        self.assertEqual(result.dtype, np.float64)

    def test_pipeline_transform(self) -> None:
        """Test element counting in a pipeline context."""
        pipeline = Pipeline(
            [
                ("smi2mol", SmilesToMol()),
                (
                    "element_count",
                    MolToElementCount(
                        element_list=[1, 6, 7, 8, 16, 35],
                        standardizer=None,
                    ),
                ),
            ],
        )
        smiles = ["CCO", "CCN", "c1ccccc1", "CS(=O)C", "CBr", "c1cc(Br)ccc1S"]
        result = pipeline.fit_transform(smiles)
        self.assertEqual(result.shape, (6, 6))
        expected = np.array(
            [
                #    H   C   N   O   S  Br
                [6.0, 2.0, 0.0, 1.0, 0.0, 0.0],  # CCO: 6H, 2C, 1O
                [7.0, 2.0, 1.0, 0.0, 0.0, 0.0],  # CCN: 7H, 2C, 1N
                [6.0, 6.0, 0.0, 0.0, 0.0, 0.0],  # benzene: 6H, 6C
                [6.0, 2.0, 0.0, 1.0, 1.0, 0.0],  # DMSO: 6H, 2C, 1O, 1S
                [3.0, 1.0, 0.0, 0.0, 0.0, 1.0],  # CBr: 3H, 1C, 1Br
                [5.0, 6.0, 0.0, 0.0, 1.0, 1.0],  # 4-bromothiophenol: 5H, 6C, 1S, 1Br
            ],
        )
        self.assertTrue(np.array_equal(result, expected))

    def test_get_set_params(self) -> None:
        """Test get_params and set_params round-trip.

        Raises
        ------
        AssertionError
            If pretransform_single does not return an np.ndarray.

        """
        element_count = MolToElementCount(element_list=[1, 6], standardizer=None)
        params = element_count.get_params(deep=True)
        self.assertEqual(params["element_list"], [1, 6])
        # Deep copy: modifying returned params should not affect original
        params["element_list"].append(7)
        self.assertEqual(element_count.element_list, [1, 6])
        count1 = element_count.pretransform_single(MolFromSmiles("CCO"))
        if not isinstance(count1, np.ndarray):
            raise AssertionError(f"Expected np.ndarray, got {type(count1)}")
        self.assertTrue(np.array_equal(count1, [6.0, 2.0]))

        # set_params should update element_list and feature_names
        element_count.set_params(element_list=[8, 16])
        self.assertEqual(element_count.element_list, [8, 16])
        self.assertEqual(element_count.feature_names, ["Count_O", "Count_S"])
        count2 = element_count.pretransform_single(MolFromSmiles("CCO"))
        if not isinstance(count2, np.ndarray):
            raise AssertionError(f"Expected np.ndarray, got {type(count2)}")
        self.assertTrue(np.array_equal(count2, [1.0, 0.0]))  # 1 O, 0 S

    def test_recreate_from_params(self) -> None:
        """Test that the element can be recreated from its params.

        Raises
        ------
        AssertionError
            If pretransform_single does not return a np.ndarray.

        """
        original = MolToElementCount(element_list=[6, 7, 8], standardizer=None)
        counts_original = original.pretransform_single(MolFromSmiles("CCO"))
        if not isinstance(counts_original, np.ndarray):
            raise AssertionError(f"Expected np.ndarray, got {type(counts_original)}")
        recreated = MolToElementCount(**original.get_params())
        self.assertEqual(original.element_list, recreated.element_list)
        self.assertEqual(original.feature_names, recreated.feature_names)
        counts_recreated = recreated.pretransform_single(MolFromSmiles("CCO"))
        if not isinstance(counts_recreated, np.ndarray):
            raise AssertionError(f"Expected np.ndarray, got {type(counts_recreated)}")
        self.assertTrue(np.array_equal(counts_original, counts_recreated))

    def test_element_list_setter(self) -> None:
        """Test that the element_list setter validates and updates feature names."""
        element_count = MolToElementCount(element_list=[1], standardizer=None)
        self.assertEqual(element_count.feature_names, ["Count_H"])
        self.assertEqual(element_count.n_features, 1)
        element_count.element_list = [6, 7]
        self.assertEqual(element_count.feature_names, ["Count_C", "Count_N"])
        self.assertEqual(element_count.n_features, 2)

        with self.assertRaises(ValueError):
            element_count.element_list = [0]

    def test_element_list_getter_returns_copy(self) -> None:
        """Test that the getter returns a copy, not the internal list."""
        element_count = MolToElementCount(element_list=[1, 6], standardizer=None)
        returned = element_count.element_list
        returned.append(99)
        self.assertEqual(element_count.element_list, [1, 6])

    def test_sklearn_clone(self) -> None:
        """Test that the element can be cloned with sklearn's clone.

        Raises
        ------
        AssertionError
            If sklearn's clone does not produce an object of same type. Only for
            typecheckers.

        """
        original = MolToElementCount(element_list=[6, 7, 8], standardizer=None)
        counts_original = original.pretransform_single(MolFromSmiles("CCO"))
        if not isinstance(counts_original, np.ndarray):
            raise AssertionError(f"Expected np.ndarray, got {type(counts_original)}")
        cloned = clone(original)
        self.assertIsNot(original, cloned)
        if not isinstance(cloned, type(original)):
            raise AssertionError(f"Expected type {type(original)}, got {type(cloned)}")
        self.assertIsNot(original.element_list, cloned.element_list)
        self.assertIsNot(original.feature_names, cloned.feature_names)
        counts_recreated = cloned.pretransform_single(MolFromSmiles("CCO"))
        if not isinstance(counts_recreated, np.ndarray):
            raise AssertionError(f"Expected np.ndarray, got {type(counts_recreated)}")
        self.assertTrue(np.array_equal(counts_original, counts_recreated))

    def test_hydrogen_counting(self) -> None:
        """Test that hydrogen counting works correctly."""
        pipeline = Pipeline(
            [
                ("smi2mol", SmilesToMol()),
                (
                    "element_count",
                    MolToElementCount(
                        element_list=[1],
                        standardizer=None,
                    ),
                ),
            ],
        )
        smiles = ["C", "[H]C(-[H])(-[H])-[H]", "[2H]C(-[2H])(-[2H])-[2H]"]
        result = pipeline.fit_transform(smiles)
        self.assertEqual(result.shape, (3, 1))
        expected = np.array(
            [
                [4.0],  # methane: 4 H
                [4.0],  # methane with explicit hydrogens: 4 H
                [4.0],  # deuterated methane: 4 H
            ],
        )
        self.assertTrue(np.array_equal(result, expected))


if __name__ == "__main__":
    unittest.main()



"""Unit tests for Gaussian process kernels using the Tanimoto similarity."""

import abc
import unittest

import numpy as np
import numpy.typing as npt
from sklearn.gaussian_process import GaussianProcessClassifier, GaussianProcessRegressor
from sklearn.gaussian_process.kernels import RBF, Kernel, Sum

from molpipeline import Pipeline
from molpipeline.any2mol import AutoToMol
from molpipeline.kernel.gaussian_process_kernel import (
    ExponentialTanimotoKernel,
    TanimotoKernel,
)
from molpipeline.kernel.tanimoto_functions import tanimoto_similarity_sparse
from molpipeline.mol2any import MolToMorganFP


class GPKernelTestMixin(abc.ABC):
    """Test case for Gaussian process kernels using Tanimoto similarity."""

    kernel: Kernel
    feature_matrix_a: npt.NDArray[np.int_]
    feature_matrix_b: npt.NDArray[np.int_]
    smiles_list: list[str]
    label: list[int]

    def setUp(self) -> None:
        """Create small integer fingerprint matrices."""
        self.feature_matrix_a = np.array(
            [
                [1, 0, 1, 0],
                [1, 1, 0, 0],
                [1, 1, 1, 0],
            ],
            dtype=int,
        )
        self.feature_matrix_b = np.array(
            [
                [1, 0, 1, 1],
                [1, 1, 0, 0],
            ],
            dtype=int,
        )

        self.smiles_list = [
            "CCO",  # Ethanol
            "CCN",  # Ethylamine
            "CCC",  # Propane
        ]
        self.label = [
            1,
            1,
            0,  # Binary label (contains hetero atom or not) for testing
        ]

    def test_addition_with_rbf(self) -> None:
        """Test addition of TanimotoKernel with RBF kernel."""
        combined_kernel = self.kernel + RBF(length_scale=1.0)
        self.assertIsInstance(combined_kernel, Sum)  # type: ignore

    def build_classification_pipeline(self) -> Pipeline:
        """Build a simple pipeline for testing.

        Returns
        -------
        Pipeline
            A Pipeline with AutoToMol, MolToMorganFP, and GaussianProcessClassifier

        """
        return Pipeline(
            [
                ("smiles2mol", AutoToMol()),
                (
                    "fingerprint",
                    MolToMorganFP(radius=2, n_bits=1024, return_as="dense"),
                ),
                ("gp_classifier", GaussianProcessClassifier(kernel=self.kernel)),
            ],
        )

    def build_regression_pipeline(self) -> Pipeline:
        """Build a simple regression pipeline for testing.

        Returns
        -------
        Pipeline
            A Pipeline with AutoToMol, MolToMorganFP, and GaussianProcessRegressor

        """
        return Pipeline(
            [
                ("smiles2mol", AutoToMol()),
                (
                    "fingerprint",
                    MolToMorganFP(radius=2, n_bits=1024, return_as="dense"),
                ),
                ("gp_regressor", GaussianProcessRegressor(kernel=self.kernel)),
            ],
        )

    def test_classification_pipeline_fit_and_predict(self) -> None:
        """Test fitting the pipeline with sample data."""
        pipeline = self.build_classification_pipeline()
        pipeline.fit(self.smiles_list, self.label)
        predictions = pipeline.predict(self.smiles_list)
        proba = pipeline.predict_proba(self.smiles_list)
        self.assertEqual(len(predictions), len(self.smiles_list))  # type: ignore
        self.assertEqual(len(proba), len(self.smiles_list))  # type: ignore

    def test_regression_pipeline_fit_and_predict(self) -> None:
        """Test fitting the regression pipeline with sample data."""
        pipeline = self.build_regression_pipeline()
        pipeline.fit(self.smiles_list, self.label)
        predictions = pipeline.predict(self.smiles_list)
        self.assertEqual(len(predictions), len(self.smiles_list))  # type: ignore
        predictions, std = pipeline.predict(self.smiles_list, return_std=True)
        self.assertEqual(len(std), len(self.smiles_list))  # type: ignore


class TestTanimotoKernel(GPKernelTestMixin, unittest.TestCase):
    """Tests for TanimotoKernel."""

    def setUp(self) -> None:
        """Create small integer fingerprint matrices."""
        super().setUp()
        self.kernel = TanimotoKernel()

    def test_call_self_no_gradient(self) -> None:
        """Test similarity matrix when Y is None (self-similarity)."""
        kernel_matrix = self.kernel(self.feature_matrix_a)
        expected = tanimoto_similarity_sparse(
            self.feature_matrix_a,
            self.feature_matrix_a,
        )
        self.assertTrue(np.allclose(kernel_matrix, expected))

    def test_call_with_y_and_gradient(self) -> None:
        """Test call with external Y and gradient path."""
        kernel_matrix, grad = self.kernel(
            self.feature_matrix_a,
            self.feature_matrix_b,
            eval_gradient=True,
        )
        kernel_matrix_expected = tanimoto_similarity_sparse(
            self.feature_matrix_a,
            self.feature_matrix_b,
        )
        self.assertTrue(np.allclose(kernel_matrix, kernel_matrix_expected))
        self.assertEqual(grad.shape, (*kernel_matrix.shape, 0))

    def test_diag(self) -> None:
        """Test diag returns vector of ones."""
        d = self.kernel.diag(self.feature_matrix_a)
        np.testing.assert_array_equal(d, np.ones(self.feature_matrix_a.shape[0]))


class TestExponentialTanimotoKernel(GPKernelTestMixin, unittest.TestCase):
    """Tests for ExponentialTanimotoKernel."""

    def setUp(self) -> None:
        """Create fingerprints without zero pairwise similarity."""
        # Ensure no zero similarities: cumulative feature inclusion
        super().setUp()
        self.kernel = ExponentialTanimotoKernel(exponent=2.0)

    def test_repr(self) -> None:
        """Test string representation includes exponent."""
        rep = repr(self.kernel)
        self.assertIn("ExponentialTanimotoKernel", rep)
        self.assertIn("exponent=2", rep)

    def test_hyperparameter_exponent_property(self) -> None:
        """Test hyperparameter_exponent exposes correct bounds."""
        hp = self.kernel.hyperparameter_exponent
        self.assertEqual(hp.name, "exponent")
        self.assertTrue(np.allclose(hp.bounds, [[1e-3, 5]]))

    def test_call_no_gradient(self) -> None:
        """Test kernel matrix without gradient equals powered similarity."""
        kernel_matrix = self.kernel(self.feature_matrix_a)

        tanimoto = tanimoto_similarity_sparse(
            self.feature_matrix_a,
            self.feature_matrix_a,
        )
        kernel_matrix_expected = tanimoto**self.kernel.exponent

        self.assertTrue(np.allclose(kernel_matrix, kernel_matrix_expected))
        # Exponent effect: off-diagonal value decreases (since exponent>1)
        self.assertTrue(
            np.all(
                kernel_matrix[np.triu_indices_from(kernel_matrix, 1)]
                < tanimoto[np.triu_indices_from(tanimoto, 1)],
            ),
        )

    def test_call_with_gradient(self) -> None:
        """Test kernel matrix and gradient computation."""
        tanimoto = tanimoto_similarity_sparse(
            self.feature_matrix_a,
            self.feature_matrix_a,
        )
        kernel_matrix, grad = self.kernel(self.feature_matrix_a, eval_gradient=True)
        # Expected sim includes small diagonal epsilon (1e-9)
        expected_sim = (
            tanimoto**self.kernel.exponent
            + np.eye(self.feature_matrix_a.shape[0]) * 1e-9
        )
        self.assertTrue(np.allclose(kernel_matrix, expected_sim))
        # Gradient shape
        self.assertEqual(
            grad.shape,
            (self.feature_matrix_a.shape[0], self.feature_matrix_a.shape[0], 1),
        )
        base_grad = (tanimoto**self.kernel.exponent) * np.log10(tanimoto)
        expected_grad = base_grad + np.eye(self.feature_matrix_a.shape[0]) * 1e-9
        self.assertTrue(np.allclose(grad[:, :, 0], expected_grad))
        self.assertFalse(np.isnan(grad).any(), "Gradient contains NaN values.")

    def test_diag(self) -> None:
        """Test diag returns ones."""
        d = self.kernel.diag(self.feature_matrix_a)
        np.testing.assert_array_equal(d, np.ones(self.feature_matrix_a.shape[0]))



"""Tests for SDF to mol."""

import unittest

from rdkit import Chem

from molpipeline.abstract_pipeline_elements.core import InvalidInstance
from molpipeline.any2mol.sdf2mol import SDFToMol


class TestSDFToMol(unittest.TestCase):
    """Test class for SDFToMol."""

    def setUp(self) -> None:
        """Set up test fixtures."""
        self.sdf_str_benzaldehyde = """240
  -OEChem-05082503512D

 14 14  0     0  0  0  0  0  0999 V2000
    3.7321    1.7500    0.0000 O   0  0  0  0  0  0  0  0  0  0  0  0
    2.8660    0.2500    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    2.0000   -0.2500    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    3.7321   -0.2500    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    2.0000   -1.2500    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    3.7321   -1.2500    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    2.8660   -1.7500    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    2.8660    1.2500    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    1.4631    0.0600    0.0000 H   0  0  0  0  0  0  0  0  0  0  0  0
    4.2690    0.0600    0.0000 H   0  0  0  0  0  0  0  0  0  0  0  0
    1.4631   -1.5600    0.0000 H   0  0  0  0  0  0  0  0  0  0  0  0
    4.2690   -1.5600    0.0000 H   0  0  0  0  0  0  0  0  0  0  0  0
    2.8660   -2.3700    0.0000 H   0  0  0  0  0  0  0  0  0  0  0  0
    2.3291    1.5600    0.0000 H   0  0  0  0  0  0  0  0  0  0  0  0
  1  8  2  0  0  0  0
  2  3  2  0  0  0  0
  2  4  1  0  0  0  0
  2  8  1  0  0  0  0
  3  5  1  0  0  0  0
  3  9  1  0  0  0  0
  4  6  2  0  0  0  0
  4 10  1  0  0  0  0
  5  7  2  0  0  0  0
  5 11  1  0  0  0  0
  6  7  1  0  0  0  0
  6 12  1  0  0  0  0
  7 13  1  0  0  0  0
  8 14  1  0  0  0  0
M  END
> <PUBCHEM_COMPOUND_CID>
240

> <PUBCHEM_COMPOUND_CANONICALIZED>
1

> <PUBCHEM_CACTVS_COMPLEXITY>
72.5

$$$$
"""

        self.sdf_str_aspirin = """2244
  -OEChem-05082504263D

 21 21  0     0  0  0  0  0  0999 V2000
    1.2333    0.5540    0.7792 O   0  0  0  0  0  0  0  0  0  0  0  0
   -0.6952   -2.7148   -0.7502 O   0  0  0  0  0  0  0  0  0  0  0  0
    0.7958   -2.1843    0.8685 O   0  0  0  0  0  0  0  0  0  0  0  0
    1.7813    0.8105   -1.4821 O   0  0  0  0  0  0  0  0  0  0  0  0
   -0.0857    0.6088    0.4403 C   0  0  0  0  0  0  0  0  0  0  0  0
   -0.7927   -0.5515    0.1244 C   0  0  0  0  0  0  0  0  0  0  0  0
   -0.7288    1.8464    0.4133 C   0  0  0  0  0  0  0  0  0  0  0  0
   -2.1426   -0.4741   -0.2184 C   0  0  0  0  0  0  0  0  0  0  0  0
   -2.0787    1.9238    0.0706 C   0  0  0  0  0  0  0  0  0  0  0  0
   -2.7855    0.7636   -0.2453 C   0  0  0  0  0  0  0  0  0  0  0  0
   -0.1409   -1.8536    0.1477 C   0  0  0  0  0  0  0  0  0  0  0  0
    2.1094    0.6715   -0.3113 C   0  0  0  0  0  0  0  0  0  0  0  0
    3.5305    0.5996    0.1635 C   0  0  0  0  0  0  0  0  0  0  0  0
   -0.1851    2.7545    0.6593 H   0  0  0  0  0  0  0  0  0  0  0  0
   -2.7247   -1.3605   -0.4564 H   0  0  0  0  0  0  0  0  0  0  0  0
   -2.5797    2.8872    0.0506 H   0  0  0  0  0  0  0  0  0  0  0  0
   -3.8374    0.8238   -0.5090 H   0  0  0  0  0  0  0  0  0  0  0  0
    3.7290    1.4184    0.8593 H   0  0  0  0  0  0  0  0  0  0  0  0
    4.2045    0.6969   -0.6924 H   0  0  0  0  0  0  0  0  0  0  0  0
    3.7105   -0.3659    0.6426 H   0  0  0  0  0  0  0  0  0  0  0  0
   -0.2555   -3.5916   -0.7337 H   0  0  0  0  0  0  0  0  0  0  0  0
  1  5  1  0  0  0  0
  1 12  1  0  0  0  0
  2 11  1  0  0  0  0
  2 21  1  0  0  0  0
  3 11  2  0  0  0  0
  4 12  2  0  0  0  0
  5  6  1  0  0  0  0
  5  7  2  0  0  0  0
  6  8  2  0  0  0  0
  6 11  1  0  0  0  0
  7  9  1  0  0  0  0
  7 14  1  0  0  0  0
  8 10  1  0  0  0  0
  8 15  1  0  0  0  0
  9 10  2  0  0  0  0
  9 16  1  0  0  0  0
 10 17  1  0  0  0  0
 12 13  1  0  0  0  0
 13 18  1  0  0  0  0
 13 19  1  0  0  0  0
 13 20  1  0  0  0  0
M  END
> <PUBCHEM_COMPOUND_CID>
2244

> <PUBCHEM_CONFORMER_RMSD>
0.6

$$$$
"""
        self.sdf_str_benzaldehyde_aspirin = (
            self.sdf_str_benzaldehyde + self.sdf_str_aspirin
        )
        self.invalid_sdf = "Not an SDF string"

    def test_initialization(self) -> None:
        """Test initialization of SDFToMol."""
        sdf2mol = SDFToMol(identifier="smiles", name="CustomName", n_jobs=2)
        self.assertEqual(sdf2mol.identifier, "smiles")
        self.assertEqual(sdf2mol.name, "CustomName")
        self.assertEqual(sdf2mol.n_jobs, 2)

    def test_pretransform_valid_sdf(self) -> None:
        """Test transformation of valid SDF string.

        Raises
        ------
        AssertionError
            If the transformation does not return a valid molecule.

        """
        sdf2mol = SDFToMol()
        # test sdf string with single molecule: benzaldehyde
        mol = sdf2mol.pretransform_single(self.sdf_str_benzaldehyde)
        self.assertIsInstance(mol, Chem.Mol)
        if not isinstance(mol, Chem.Mol):
            # necessary for mypy
            raise AssertionError("Expected a Chem.Mol object")
        self.assertEqual(mol.GetNumAtoms(), 8)

        # test sdf string with single molecule: aspirin
        mol = sdf2mol.pretransform_single(self.sdf_str_aspirin)
        self.assertIsInstance(mol, Chem.Mol)
        if not isinstance(mol, Chem.Mol):
            # necessary for mypy
            raise AssertionError("Expected a Chem.Mol object")
        self.assertEqual(mol.GetNumAtoms(), 13)

        # test sdf string with multiple molecules: benzaldehyde and aspirin
        # the current behavior is to return the first molecule only
        mol = sdf2mol.pretransform_single(self.sdf_str_benzaldehyde_aspirin)
        self.assertIsInstance(mol, Chem.Mol)
        if not isinstance(mol, Chem.Mol):
            # necessary for mypy
            raise AssertionError("Expected a Chem.Mol object")
        self.assertEqual(mol.GetNumAtoms(), 8)

    def test_pretransform_invalid_sdf(self) -> None:
        """Test handling of invalid SDF input."""
        result = SDFToMol().pretransform_single(self.invalid_sdf)
        self.assertIsInstance(result, InvalidInstance)

    def test_transform(self) -> None:
        """Test transform function.

        Raises
        ------
        AssertionError
            If the transformation does not return a valid molecule.

        """
        # test list of sdf strings
        mols = SDFToMol().transform([self.sdf_str_benzaldehyde, self.sdf_str_aspirin])
        self.assertEqual(len(mols), 2)
        self.assertIsInstance(mols[0], Chem.Mol)
        self.assertIsInstance(mols[1], Chem.Mol)
        if not isinstance(mols[0], Chem.Mol):
            # necessary for mypy
            raise AssertionError("Expected a Chem.Mol object")
        self.assertEqual(mols[0].GetNumAtoms(), 8)
        if not isinstance(mols[1], Chem.Mol):
            # necessary for mypy
            raise AssertionError("Expected a Chem.Mol object")
        self.assertEqual(mols[1].GetNumAtoms(), 13)

        # test single sdf string with multiple molecules
        # only the first molecule in the SDF string will be read
        mols = SDFToMol().transform([self.sdf_str_benzaldehyde_aspirin])
        self.assertEqual(len(mols), 1)
        self.assertIsInstance(mols[0], Chem.Mol)
        if not isinstance(mols[0], Chem.Mol):
            # necessary for mypy
            raise AssertionError("Expected a Chem.Mol object")
        self.assertEqual(mols[0].GetNumAtoms(), 8)

        # test multiple sdf strings with multiple molecules
        # the current behavior is to return the first molecule only
        mols = SDFToMol().transform(
            [self.sdf_str_benzaldehyde_aspirin, self.sdf_str_benzaldehyde],
        )
        self.assertEqual(len(mols), 2)
        self.assertIsInstance(mols[0], Chem.Mol)
        self.assertIsInstance(mols[1], Chem.Mol)
        if not isinstance(mols[0], Chem.Mol):
            # necessary for mypy
            raise AssertionError("Expected a Chem.Mol object")
        self.assertEqual(mols[0].GetNumAtoms(), 8)
        if not isinstance(mols[1], Chem.Mol):
            # necessary for mypy
            raise AssertionError("Expected a Chem.Mol object")
        self.assertEqual(mols[1].GetNumAtoms(), 8)

    def test_sdf_properties_transfer(self) -> None:
        """Test that properties from SDF are transferred to molecule.

        Raises
        ------
        AssertionError
            If the transformation does not return a valid molecule.

        """
        mol = SDFToMol().pretransform_single(self.sdf_str_benzaldehyde)
        if not isinstance(mol, Chem.Mol):
            # necessary for mypy
            raise AssertionError("Expected a Chem.Mol object")
        self.assertTrue(mol.HasProp("PUBCHEM_COMPOUND_CID"))
        self.assertEqual(mol.GetProp("PUBCHEM_COMPOUND_CID"), "240")
        self.assertTrue(mol.HasProp("PUBCHEM_COMPOUND_CANONICALIZED"))
        self.assertEqual(mol.GetProp("PUBCHEM_COMPOUND_CANONICALIZED"), "1")
        self.assertTrue(mol.HasProp("PUBCHEM_CACTVS_COMPLEXITY"))
        self.assertEqual(mol.GetProp("PUBCHEM_CACTVS_COMPLEXITY"), "72.5")



"""Legacy import path for kernel functions."""

from loguru import logger

from molpipeline.kernel.tanimoto_functions import (
    self_tanimoto_distance,
    self_tanimoto_similarity,
    tanimoto_distance_sparse,
    tanimoto_similarity_sparse,
)

logger.warning(
    "The kernel functions have been moved to molpipeline.kernel.tanimoto_functions and "
    "will be removed from utils.kernel in future releases.",
)

__all__ = [
    "self_tanimoto_distance",
    "self_tanimoto_similarity",
    "tanimoto_distance_sparse",
    "tanimoto_similarity_sparse",
]



"""Implementation of MACCS key fingerprint."""

import numpy as np
from numpy import typing as npt
from rdkit.Chem import MACCSkeys
from rdkit.DataStructs import ExplicitBitVect

from molpipeline.abstract_pipeline_elements.mol2any.mol2bitvector import (
    FPReturnAsOption,
    MolToFingerprintPipelineElement,
)
from molpipeline.utils.molpipeline_types import RDKitMol


class MolToMACCSFP(MolToFingerprintPipelineElement):
    """MACCS key fingerprint.

    The MACCS keys are a set of 166 keys that encode the presence or absence of
    particular substructures in a molecule. The MACCS keys are a subset of the
    PubChem substructure keys.

    """

    _n_bits = 167  # MACCS keys have 166 bits + 1 bit for an all-zero vector (bit 0)

    def __init__(
        self,
        return_as: FPReturnAsOption = "sparse",
        name: str = "MolToMACCSFP",
        n_jobs: int = 1,
        uuid: str | None = None,
    ) -> None:
        """Initialize MolToMACCSFP.

        Parameters
        ----------
        return_as : FPReturnAsOption, default="sparse"
            Type of output. When "sparse" the fingerprints will be returned as a
            scipy.sparse.csr_matrix holding a sparse representation of the bit vectors.
            With "dense" a numpy matrix will be returned. With "rdkit" the fingerprints
            will be returned as a list of one of RDKit's ExplicitBitVect,
            IntSparseBitVect, UIntSparseBitVect, etc. depending on the fingerprint
            and parameters.
        name : str, default="MolToMACCSFP"
            Name of PipelineElement.
        n_jobs : int, default=1
            Number of cores to use.
        uuid : str | None, optional
            UUID of the PipelineElement.

        """
        super().__init__(  # pylint: disable=R0801
            return_as=return_as,
            name=name,
            n_jobs=n_jobs,
            uuid=uuid,
        )
        self._feature_names = [f"maccs_{i}" for i in range(self._n_bits)]

    def pretransform_single(
        self,
        value: RDKitMol,
    ) -> dict[int, int] | npt.NDArray[np.int_] | ExplicitBitVect:
        """Transform a single molecule to MACCS key fingerprint.

        Parameters
        ----------
        value : RDKitMol
            RDKit molecule.

        Returns
        -------
        dict[int, int] | npt.NDArray[np.int_] | ExplicitBitVect
            MACCS key fingerprint.

        Raises
        ------
        ValueError
            If the variable `return_as` is not one of the allowed values.

        """
        fingerprint = MACCSkeys.GenMACCSKeys(value)  # type: ignore[attr-defined]
        if self._return_as == "rdkit":
            return fingerprint
        if self._return_as == "dense":
            return np.array(fingerprint)
        if self._return_as == "sparse":
            return dict.fromkeys(fingerprint.GetOnBits(), 1)
        raise ValueError(f"Unknown return_as value: {self._return_as}")



"""Definition of types used in molpipeline."""

from collections.abc import Sequence
from numbers import Number
from typing import (
    Any,
    Literal,
    Protocol,
    Self,
    TypeAlias,
    TypeVar,
)

import numpy as np
import numpy.typing as npt
from scipy.sparse import coo_matrix, csc_matrix, csr_matrix, spmatrix

from molpipeline.abstract_pipeline_elements.core import (
    ABCPipelineElement,
    OptionalMol,
    RDKitMol,
)
from molpipeline.error_handling import FilterReinserter

__all__ = [
    "AnyElement",
    "AnyNumpyElement",
    "AnyPredictor",
    "AnySklearnEstimator",
    "AnyStep",
    "AnyTransformer",
    "Number",
    "OptionalMol",
    "RDKitMol",
    "XVar",
    "YVar",
]
# One-liner type definitions

AnyNumpyElement = TypeVar("AnyNumpyElement", bound=np.generic)

_T = TypeVar("_T")
_NT = TypeVar("_NT", bound=np.generic)
TypeFixedVarSeq = TypeVar("TypeFixedVarSeq", bound=Sequence[_T] | npt.NDArray[_NT])  # type: ignore
AnyVarSeq = TypeVar("AnyVarSeq", bound=Sequence[Any] | npt.NDArray[Any])

SparseMatrix = csc_matrix[Any] | coo_matrix[Any] | csr_matrix[Any]
XType = npt.ArrayLike | npt.NDArray[Any] | spmatrix  # Generic model input features
YType = npt.ArrayLike | npt.NDArray[Any] | None  # Generic model target values
# XVar indicates that the function accepts multiple types, and returns the same type
# as the input. e.g. row removal or value manipulations.
XVar = TypeVar("XVar", bound=npt.ArrayLike | npt.NDArray[Any] | spmatrix)
# Same as XVar but for target values.
YVar = TypeVar("YVar", bound=npt.ArrayLike | npt.NDArray[Any] | None)

# FloatCountRange needs renaming to FloatRange
FloatCountRange: TypeAlias = tuple[float | None, float | None]
# IntCountRange needs renaming to IntRange
IntCountRange: TypeAlias = tuple[int | None, int | None]

# IntOrIntCountRange for Typing of count ranges
# - a single int for an exact value match
# - a range given as a tuple with a lower and upper bound
#   - both limits are optional
IntOrIntCountRange: TypeAlias = int | IntCountRange


class AnySklearnEstimator(Protocol):
    """Protocol for sklearn estimators."""

    def get_params(self, deep: bool = True) -> dict[str, Any]:
        """Get parameters for this estimator.

        Parameters
        ----------
        deep: bool
            If True, will return the parameters for this estimator.

        Returns
        -------
        dict[str, Any]
            Parameter names mapped to their values.

        """
        ...  # pylint: disable=unnecessary-ellipsis

    def set_params(self, **params: Any) -> Self:
        """Set the parameters of this estimator.

        Parameters
        ----------
        params: Any
            Estimator parameters.

        Returns
        -------
        Self
            Estimator with updated parameters.

        """
        ...  # pylint: disable=unnecessary-ellipsis

    def fit(
        self,
        X: XType,  # noqa: N803
        y: YType,
        **fit_params: Any,
    ) -> Self:
        """Fit the model with X.

        Parameters
        ----------
        X: npt.NDArray[Any]
            Model input.
        y: npt.NDArray[Any] | None
            Target values.
        fit_params: Any
            Additional parameters for fitting.


        Returns
        -------
        Self
            Fitted estimator.

        """
        ...  # pylint: disable=unnecessary-ellipsis


class AnyPredictor(AnySklearnEstimator, Protocol):
    """Protocol for predictors."""

    def fit_predict(
        self,
        X: XType,  # noqa: N803
        y: YType,
        **fit_params: Any,
    ) -> npt.NDArray[Any]:
        """Fit the model with X and return predictions.

        Parameters
        ----------
        X: npt.NDArray[Any]
            Model input.
        y: npt.NDArray[Any] | None
            Target values.
        fit_params: Any
            Additional parameters for fitting.

        Returns
        -------
        npt.NDArray[Any]
            Predictions.

        """
        ...  # pylint: disable=unnecessary-ellipsis


class AnyTransformer(AnySklearnEstimator, Protocol):
    """Protocol for transformers."""

    def fit_transform(
        self,
        X: npt.NDArray[Any],  # noqa: N803
        y: npt.NDArray[Any] | None,
        **fit_params: Any,
    ) -> npt.NDArray[Any]:
        """Fit the model with X and return the transformed array.

        Parameters
        ----------
        X: npt.NDArray[Any]
            Model input.
        y: npt.NDArray[Any] | None
            Target values.
        fit_params: Any
            Additional parameters for fitting.


        Returns
        -------
        npt.NDArray[Any]
            Transformed array.

        """
        ...  # pylint: disable=unnecessary-ellipsis

    def transform(
        self,
        X: npt.NDArray[Any],  # noqa: N803
        **params: Any,
    ) -> npt.NDArray[Any]:
        """Transform and return X according to object protocol.

        Parameters
        ----------
        X: npt.NDArray[Any]
            Model input.
        params: Any
            Additional parameters for transforming.

        Returns
        -------
        npt.NDArray[Any]
            Transformed array.

        """
        ...  # pylint: disable=unnecessary-ellipsis


AnyElement = (
    AnyTransformer
    | AnyPredictor
    | ABCPipelineElement
    | Literal["passthrough"]
    | FilterReinserter[Any]
)
AnyStep = tuple[str, AnyElement]
